"""
Test the CastToDatatype cleansing function
"""

import datetime
from decimal import Decimal

import pytest

from pydantic import ValidationError

from pyspark.sql import DataFrame
from pyspark.sql import functions as f

from koheesio.logger import LoggingFactory
from koheesio.spark.transformations.cast_to_datatype import (
    CastToBinary,
    CastToBoolean,
    CastToByte,
    CastToDatatype,
    CastToDecimal,
    CastToDouble,
    CastToFloat,
    CastToInteger,
    CastToLong,
    CastToShort,
    CastToString,
    CastToTimestamp,
)
from koheesio.spark.utils import SparkDatatype

pytestmark = pytest.mark.spark


@pytest.mark.parametrize(
    "input_values,expected",
    [
        (
            # description: Test CastToDatatype - byte to int
            # input values
            dict(column="byte", datatype="integer", target_column="integer"),
            # expected output
            1,
        ),
        (
            # description: Test CastToDatatype - short to int
            # input values
            dict(column="short", datatype="integer", target_column="integer"),
            # expected output
            1,
        ),
        (
            # description: Test CastToDatatype - integer to string
            # input values
            dict(column="integer", datatype="string", target_column="string"),
            # expected output
            "1",
        ),
        (
            # description: Test CastToDatatype - long to string
            # input values
            dict(column="long", datatype="string", target_column="string"),
            # expected output
            "1",
        ),
        (
            # description: Test CastToDatatype - float to string
            # input values
            dict(column="float", datatype="string", target_column="string"),
            # expected output
            "1.0",
        ),
        (
            # description: Test CastToDatatype - double to string
            # input values
            dict(column="double", datatype="string", target_column="string"),
            # expected output
            "1.0",
        ),
        (
            # description: Test CastToDatatype - decimal to string
            # input values
            dict(column="decimal", datatype="integer", target_column="integer"),
            # expected output
            1,
        ),
        (
            # description: Test CastToDatatype - boolean to string
            # input values
            dict(column="boolean", datatype="string", target_column="string"),
            # expected output
            "true",
        ),
        (
            # description: Test CastToDatatype - date to string
            # input values
            dict(column="date", datatype="string", target_column="string"),
            # expected output
            "2023-01-01",
        ),
        (
            # description: Test CastToDatatype - timestamp to string
            # input values
            dict(column="timestamp", datatype="string", target_column="string"),
            # expected output
            "2023-01-01 00:01:01",
        ),
        (
            # description: Test CastToDatatype - string to boolean
            # input values
            dict(column="string", datatype="boolean", target_column="boolean"),
            # expected output
            None,
        ),
        (
            # description: Test CastToDatatype - void to string
            # input values
            dict(column="void", datatype="string", target_column="string"),
            # expected output
            None,
        ),
        (
            # description: Test CastToDatatype - array to string
            # input values
            dict(column="array", datatype="string", target_column="string"),
            # expected output
            "[a]",
        ),
        (
            # description: Test CastToDatatype - map to string
            # input values
            dict(column="map", datatype="string", target_column="string"),
            # expected output
            "{a -> b}",
        ),
        (
            # description: Test CastToDatatype - binary to string
            # input values
            dict(column="binary", datatype="string", target_column="string"),
            # expected output
            "a",
        ),
        (
            # description: Test CastToDatatype - boolean to integer
            # input values
            dict(column="boolean", datatype="integer", target_column="integer"),
            # expected output
            1,
        ),
        (
            # description: Test CastToDatatype - using SparkDatatype enum
            # input values
            dict(column="boolean", datatype=SparkDatatype.INTEGER, target_column="integer"),
            # expected output
            1,
        ),
    ],
)
def test_happy_flow(input_values, expected, df_with_all_types: DataFrame):
    log = LoggingFactory.get_logger(name="test_cast_to_datatype")

    cast_to_datatype = CastToDatatype(**input_values)
    output_df = cast_to_datatype.transform(df_with_all_types)
    col = cast_to_datatype.columns[0]
    target_column = cast_to_datatype.target_column

    # log equivalent of doing df.show()
    log.error(f"show output_df: \n{output_df.select(col, target_column)._jdf.showString(20, 20, False)}")

    actual = [row[target_column] for row in output_df.select(target_column).collect()][0]
    assert actual == expected


def test_wrong_datatype():
    """
    Should only accept atomic types as defined in pyspark.sql.types
    """
    with pytest.raises(AttributeError):
        CastToDatatype(column="c1", datatype="sting", target_column="c1")
        CastToDatatype(column="c1", datatype=123, target_column="c1")


@pytest.mark.parametrize(
    "klass,expected",
    [
        (
            CastToByte,
            {
                "short_target": 1,
                "integer_target": 1,
                "long_target": 1,
                "float_target": 1,
                "double_target": 1,
                "decimal_target": 1,
                # 'string_target': None,
                "boolean_target": 1,
                # 'timestamp_target': 61,
                # 'date_target': None,
                # 'void_target': None,
            },
        ),
        (
            CastToShort,
            {
                "byte_target": 1,
                "integer_target": 1,
                "long_target": 1,
                "float_target": 1,
                "double_target": 1,
                "decimal_target": 1,
                # 'string_target': None,
                "boolean_target": 1,
                # 'timestamp_target': -12995,
                # 'date_target': None,
                # 'void_target': None,
            },
        ),
        (
            CastToInteger,
            {
                "byte_target": 1,
                "short_target": 1,
                "long_target": 1,
                "float_target": 1,
                "double_target": 1,
                "decimal_target": 1,
                # 'string_target': None,
                "boolean_target": 1,
                "timestamp_target": 1672531261,  # '2023-01-01T00:01:01'
                # 'date_target': None,
                # 'void_target': None,
            },
        ),
        (
            CastToLong,
            {
                "byte_target": 1,
                "short_target": 1,
                "integer_target": 1,
                "float_target": 1,
                "double_target": 1,
                "decimal_target": 1,
                # 'string_target': None,
                "boolean_target": 1,
                "timestamp_target": 1672531261,  # '2023-01-01T00:01:01'
                # 'date_target': None,
                # 'void_target': None,
            },
        ),
        (
            CastToFloat,
            {
                "byte_target": 1.0,
                "short_target": 1.0,
                "integer_target": 1.0,
                "long_target": 1.0,
                "double_target": 1.0,
                "decimal_target": 1.0,
                # 'string_target': None,
                "boolean_target": 1.0,
                # 'timestamp_target': 1.6725312E9,  # ~ '2023-01-01T00:01:01' - loss of detail with a regular Float
                # 'date_target': None,
                # 'void_target': None,
            },
        ),
        (
            CastToDouble,
            {
                "byte_target": 1.0,
                "short_target": 1.0,
                "integer_target": 1.0,
                "long_target": 1.0,
                "float_target": 1.0,
                "decimal_target": 1.0,
                # 'string_target': None,
                "boolean_target": 1.0,
                "timestamp_target": 1672531261.0,  # '2022-12-31T23:01:01'
                # 'date_target': None,
                # 'void_target': None,
            },
        ),
        (
            CastToDecimal,
            {
                "byte_target": Decimal("1.000000000000000000"),
                "short_target": Decimal("1.000000000000000000"),
                "integer_target": Decimal("1.000000000000000000"),
                "long_target": Decimal("1.000000000000000000"),
                "float_target": Decimal("1.000000000000000000"),
                "double_target": Decimal("1.000000000000000000"),
                "decimal_target": Decimal("1.000000000000000000"),
                # 'string_target': None,
                "boolean_target": Decimal("1.000000000000000000"),
                "timestamp_target": Decimal("1672531261.000000000000000000"),
                # 'date_target': None,
                # 'void_target': None,
            },
        ),
        (
            CastToString,
            {
                "byte_target": "1",
                "short_target": "1",
                "integer_target": "1",
                "long_target": "1",
                "float_target": "1.0",
                "double_target": "1.0",
                "decimal_target": "1",
                "binary_target": "a",
                "boolean_target": "true",
                "timestamp_target": "2023-01-01 00:01:01",
                "date_target": "2023-01-01",
                "array_target": "[a]",
                "map_target": "{a -> b}",
                # 'void_target': None,
            },
        ),
        (
            CastToBinary,
            {
                "byte_target": bytearray(b"\x01"),
                "short_target": bytearray(b"\x00\x01"),
                "integer_target": bytearray(b"\x00\x00\x00\x01"),
                "long_target": bytearray(b"\x00\x00\x00\x00\x00\x00\x00\x01"),
                # 'float_target': AnalysisException,
                # 'double_target': AnalysisException,
                # 'decimal_target': AnalysisException,
                "string_target": bytearray(b"a"),
                # 'boolean_target': AnalysisException,
                # 'timestamp_target': AnalysisException,
                # 'date_target': AnalysisException,
                # 'array_target': AnalysisException,
                # 'map_target': AnalysisException,
                # 'void_target': None,
            },
        ),
        (
            CastToBoolean,
            {
                "byte_target": True,
                "short_target": True,
                "integer_target": True,
                "long_target": True,
                "float_target": True,
                "double_target": True,
                "decimal_target": True,
                # 'string_target': None,
                # 'binary_target': AnalysisException,
                "timestamp_target": True,
                # 'date_target': None,
                # 'array_target': AnalysisException,
                # 'map_target': AnalysisException,
                # 'void_target': None,
            },
        ),
        (
            CastToTimestamp,
            {
                # 0 = '1970-01-01T00:00:01'
                # 'byte_target': datetime.datetime.utcfromtimestamp(1),
                # 'short_target': datetime.datetime.utcfromtimestamp(1),
                "integer_target": datetime.datetime.utcfromtimestamp(1),
                "long_target": datetime.datetime.utcfromtimestamp(1),
                "float_target": datetime.datetime.utcfromtimestamp(1),
                "double_target": datetime.datetime.utcfromtimestamp(1),
                "decimal_target": datetime.datetime.utcfromtimestamp(1),
                # 'string_target': None,
                # 'binary_target': AnalysisException,
                # 'boolean_target': datetime.datetime.utcfromtimestamp(1),
                "date_target": datetime.datetime(2023, 1, 1, 0, 0),
                # 'array_target': AnalysisException,
                # 'map_target': AnalysisException,
                # 'void_target': None,
            },
        ),
    ],
)
def test_cast_to_specific_type(klass, expected, df_with_all_types):
    log = LoggingFactory.get_logger(name="test_cast_to_specific_type")

    klz = klass(df=df_with_all_types, target_suffix="target")
    target_columns = [tc for tc, _ in klz.get_columns_with_target()]
    output = klz.execute()
    output_df = output.df.select(target_columns)
    actual = output_df.head().asDict()

    # log equivalent of doing df.show()
    log.error(f"show actual: \n{output_df._jdf.showString(20, 20, False)}")

    assert target_columns == list(expected.keys())
    assert actual == expected


@pytest.mark.parametrize(
    "precision,scale,alternative_value,expected",
    [
        # with a precision of 1, the scale can only be 0; 1 will result in a null value
        (1, 1, Decimal("1"), [{"c1": None, "c2": None}]),
        (1, 0, Decimal("42"), [{"c1": Decimal("1"), "c2": None}]),
        # with a precision of 10, the scale can be 0 - 10
        (10, 1, Decimal("123456789.0123456789"), [{"c1": Decimal("1.0"), "c2": Decimal("123456789.0")}]),
        (10, 9, Decimal("123456789.0123456789"), [{"c1": Decimal("1.0"), "c2": None}]),
        (10, 10, Decimal("1"), [{"c1": None, "c2": None}]),
        # Koheesio defaults to a precision of 38 and a scale of 18 - c2 is of max precision and scale in this case
        (
            38,
            18,
            Decimal("12345678901234567890.012345678901234567"),
            [{"c1": Decimal("1.000000000000000000"), "c2": Decimal("12345678901234567890.012345678901234567")}],
        ),
        # Unhappy flows
        # Values that are too big or too small for precision and scale will result in an error getting raised
        (-1, 0, ..., ValidationError),
        (39, 0, ..., ValidationError),
        (1, -1, ..., ValidationError),
    ],
)
def test_decimal_precision_and_scale(precision, scale, alternative_value, expected, df_with_all_types):
    log = LoggingFactory.get_logger(name="test_decimal_precision_and_scale")

    if alternative_value == ...:
        with pytest.raises(expected):
            CastToDecimal(columns=["decimal"], scale=scale, precision=precision)
        return

    input_df = (
        df_with_all_types.select("decimal")
        .withColumn("c1", f.col("decimal"))
        .withColumn("c2", f.lit(alternative_value))
        .select("c1", "c2")
    )

    # log equivalent of doing df.show() and df.printSchema()
    log.error(f"show input_df: \n{input_df._jdf.showString(20, 20, False)}")
    log.error(f"printSchema input_df: \n{input_df._jdf.schema().treeString()}")

    output_df = CastToDecimal(columns=["c1", "c2"], scale=scale, precision=precision).transform(input_df)
    log.error(f"show output_df: \n{output_df._jdf.showString(20, 20, False)}")
    log.error(f"printSchema output_df: \n{output_df._jdf.schema().treeString()}")

    actual = [row.asDict() for row in output_df.collect()]
    assert actual == expected


# TODO: nested functionality does not exists yet
#
#     def test_transform_nested_df(self):
#         """
#         Test with nested dataframes
#         """
#         output_df = (
#             CastToDatatype(
#                 df=self.sample_nested_df,
#                 params={
#                     "source_column": "payload.col2.col_int",
#                     "datatype": "string",
#                 },
#                 columns_to_flatten=["payload{}.col2{}"],
#                 target_alias="casted_int_column",
#             )
#             .execute()
#             .df
#         )
#
#         self.assertEqual(output_df.select("casted_int_column").dtypes[0][1], "string")



import logging
import os
from pathlib import Path
from unittest.mock import patch

import pytest
from conftest import setup_test_data

from pydantic import ValidationError

from pyspark.sql.types import LongType

from koheesio.logger import LoggingFactory
from koheesio.spark.delta import DeltaTableStep

pytestmark = pytest.mark.spark

log = LoggingFactory.get_logger(name="test_delta", inherit_from_koheesio=True)


@pytest.mark.parametrize(
    "value,expected",
    [
        (
            "test_catalog.test_schema.test_table",
            {
                "catalog": "test_catalog",
                "database": "test_schema",
                "table": "test_table",
                "table_name": "test_catalog.test_schema.test_table",
            },
        ),
        (
            "test_schema.test_table",
            {
                "catalog": None,
                "database": "test_schema",
                "table": "test_table",
                "table_name": "test_schema.test_table",
            },
        ),
        (
            {"catalog": "a_catalog", "table": "a_schema.a_table"},
            {
                "catalog": "a_catalog",
                "database": "a_schema",
                "table": "a_table",
                "table_name": "a_catalog.a_schema.a_table",
            },
        ),
        (
            "delta_table",
            {
                "catalog": None,
                "database": None,
                "table": "delta_table",
                "table_name": "delta_table",
            },
        ),
    ],
)
def test_table(value, expected):
    if isinstance(value, str):
        actual = DeltaTableStep(table=value)
        log.info(f"actual: {actual}")
    elif isinstance(value, dict):
        actual = DeltaTableStep(**value)
    else:
        raise AssertionError(f"Unable to run test with given input '{value}'")

    assert actual.catalog == expected["catalog"]
    assert actual.database == expected["database"]
    assert actual.table == expected["table"]
    assert actual.table_name == expected["table_name"]

    log.info("delta test completed")


def test_delta_table_properties(spark, setup, delta_file):
    setup_test_data(spark=spark, delta_file=Path(delta_file))
    table_name = "delta_test_table"
    dt = DeltaTableStep(
        table=table_name,
        default_create_properties={
            "delta.randomizeFilePrefixes": "true",
            "delta.checkpoint.writeStatsAsStruct": "true",
            "delta.minReaderVersion": "2",
            "delta.minWriterVersion": "5",
            "not_string_property_but_bool": True,
            "not_string_property_but_int": 123,
        },
    )

    assert dt.exists is True
    assert dt.dataframe.count() == 10
    assert dt.columns == ["id"]
    assert dt.get_column_type("id") == LongType()
    assert dt.get_column_type("id").simpleString() == "bigint"
    assert dt.get_column_type("none") is None
    assert dt.get_column_type(None) is None
    assert dt.has_change_type is False

    assert dt.table_name == table_name

    properties = dt.get_persisted_properties()
    assert properties.get("delta.enableChangeDataFeed") == "true"
    dt.add_properties({"delta.enableChangeDataFeed": "false"}, override=True)
    assert dt.get_persisted_properties().get("delta.enableChangeDataFeed") == "false"

    dt.add_properties({"delta.enableChangeDataFeed": "true"})
    assert dt.get_persisted_properties().get("delta.enableChangeDataFeed") == "false"

    assert properties.get("test_property") is None
    dt.add_properties({"test_property": "delta_table_value"})
    assert dt.get_persisted_properties().get("test_property") == "delta_table_value"

    assert dt.default_create_properties.get("not_string_property_but_bool") == "true"
    assert dt.default_create_properties.get("not_string_property_but_int") == "123"

    dt.add_properties({"test_property_bool": True})
    assert dt.get_persisted_properties().get("test_property_bool") == "true"
    dt.add_properties({"test_property_int": 123})
    assert dt.get_persisted_properties().get("test_property_int") == "123"

    for v in dt.default_create_properties.values():
        assert isinstance(v, str)


@patch.dict(
    os.environ,
    {
        "DATABRICKS_RUNTIME_VERSION": "lts_11_spark_3_scala_2.12",
    },
)
def test_delta_table_properties_dbx():
    dt = DeltaTableStep(table="delta_test_table_dbx", create_if_not_exists=True)
    assert dt.default_create_properties.get("delta.autoOptimize.autoCompact") == "true"


@pytest.mark.parametrize("value,expected", [("too.many.dots.given.to.be.a.real.table", pytest.raises(ValidationError))])
def test_table_failed(value, expected):
    with expected:
        DeltaTableStep(table=value)

    dt = DeltaTableStep(table="unknown_table")
    assert dt.exists is False


@pytest.mark.parametrize(
    ["table", "create_if_not_exists", "log_level"], [("unknown", False, "DEBUG"), ("unknown", True, "INFO")]
)
def test_exists(caplog, table, create_if_not_exists, log_level):
    with caplog.at_level(log_level):
        dt = DeltaTableStep(table=table, create_if_not_exists=create_if_not_exists)
        dt.log.setLevel(log_level)
        assert dt.exists is False
        assert f"The `create_if_not_exists` flag is set to {create_if_not_exists}." in caplog.text



"""
Trim whitespace from the beginning and/or end of a string.

Classes
-------
- `Trim`
    Trim whitespace from the beginning and/or end of a string.
- `LTrim`
    Trim whitespace from the beginning of a string.
- `RTrim`
    Trim whitespace from the end of a string.

See class docstrings for more information.
"""

from typing import Literal

import pyspark.sql.functions as f
from pyspark.sql import Column

from koheesio.models import Field, ListOfColumns
from koheesio.spark.transformations import ColumnsTransformationWithTarget
from koheesio.spark.utils import SparkDatatype

trim_type = Literal["left", "right", "left-right"]


class Trim(ColumnsTransformationWithTarget):
    """
    Trim whitespace from the beginning and/or end of a string.

    This is a wrapper around PySpark ltrim() and rtrim() functions

    The `direction` parameter can be changed to apply either a left or a right trim. Defaults to left AND right trim.

    Note: If the type of the column is not string, Trim will not be run. A Warning will be thrown indicating this

    Parameters
    ----------
    columns : ListOfColumns
        The column (or list of columns) to trim. Alias: column
        If no columns are provided, all string columns will be trimmed.
    target_column : ListOfColumns, optional
        The column to store the result in. If not provided, the result will be stored in the source column.
        Alias: target_suffix - if multiple columns are given as source, this will be used as a suffix.
    direction : trim_type, optional, default "left-right"
        On which side to remove the spaces. Either "left", "right" or "left-right". Defaults to "left-right"

    Examples
    --------

    __input_df:__
    | column    |
    |-----------|
    | " hello " |

    ### Trim whitespace from the beginning of a string

    ```python
    output_df = Trim(
        column="column", target_column="trimmed_column", direction="left"
    ).transform(input_df)
    ```

    __output_df:__
    | column    | trimmed_column |
    |-----------|----------------|
    | " hello " | "hello "       |


    ### Trim whitespace from both sides of a string

    ```python
    output_df = Trim(
        column="column",
        target_column="trimmed_column",
        direction="left-right",  # default value
    ).transform(input_df)
    ```

    __output_df:__
    | column    | trimmed_column |
    |-----------|----------------|
    | " hello " | "hello"        |

    ### Trim whitespace from the end of a string

    ```python
    output_df = Trim(
        column="column", target_column="trimmed_column", direction="right"
    ).transform(input_df)
    ```

    __output_df:__
    | column    | trimmed_column |
    |-----------|----------------|
    | " hello " | " hello"       |

    """

    class ColumnConfig(ColumnsTransformationWithTarget.ColumnConfig):
        """Limit data types to string only."""

        run_for_all_data_type = [SparkDatatype.STRING]
        limit_data_type = [SparkDatatype.STRING]

    columns: ListOfColumns = Field(
        default="*",
        alias="column",
        description="The column (or list of columns) to trim. Alias: column. If no columns are provided, all string"
        "columns will be trimmed.",
    )
    direction: trim_type = Field(
        default="left-right", description="On which side to remove the spaces. Either 'left', 'right' or 'left-right'"
    )

    def func(self, column: Column):
        if self.direction == "left":
            return f.ltrim(column)

        if self.direction == "right":
            return f.rtrim(column)

        # both (left-right)
        return f.rtrim(f.ltrim(column))


class LTrim(Trim):
    """Trim whitespace from the beginning of a string. Alias: LeftTrim"""

    direction: trim_type = "left"


class RTrim(Trim):
    """Trim whitespace from the end of a string. Alias: RightTrim"""

    direction: trim_type = "right"



"""
Test the Replace columns transformation
"""

import pytest

from koheesio.logger import LoggingFactory
from koheesio.spark.transformations.strings.replace import Replace

pytestmark = pytest.mark.spark

log = LoggingFactory.get_logger(name=__file__, inherit_from_koheesio=True)

data_with_strings = [[1, 2, "hello"], [3, 4, "world"], [3, 4, None]]
schema_with_strings = ["c1", "c2", "c3"]


@pytest.mark.parametrize(
    "input_values,expected",
    [
        (
            # description: Test Replace
            # input values
            dict(column="c2", original_value=2, new_value=100),
            # expected data
            ["100", "4", "4"],
        ),
        (
            # description: Test Replace - testing with None value
            # input values
            dict(column="c3", new_value="NA"),
            # expected data
            ["hello", "world", "NA"],
        ),
        (
            # description: Test Replace - string column value
            # input values
            dict(column="c3", original_value="world", new_value="programmer"),
            # expected data
            ["hello", "programmer", None],
        ),
    ],
)
def test_happy_flow(input_values, expected, spark):
    input_df = spark.createDataFrame(data_with_strings, schema_with_strings)

    replace = Replace(**input_values)
    target_column = replace.columns[0]
    output_df = replace.transform(input_df)

    # log equivalent of doing df.show()
    log.info(f"show output_df: \n{output_df._jdf.showString(20, 20, False)}")
    actual = [row.asDict()[target_column] for row in output_df.collect()]
    assert actual == expected


# TODO: not supported yet
#
#     def test_nested(self):
#         df = self.spark.createDataFrame([{"a": {"b": "hello"}}, {"a": {"b": "world"}}], self.schema)
#         df = (
#             Replace(
#                 df=df,
#                 params={
#                     "source_column": "a.b",
#                     "original_value": "world",
#                     "new_value": "programmer",
#                 },
#                 target_alias="a.c",
#                 columns_to_flatten=["a{}.b"],
#             )
#             .execute()
#             .df
#         )
#         result = df.rdd.map(lambda r: r.a.c).collect()
#         self.assertEqual(sorted(result), ["hello", "programmer"])
#
#     def test_nested_1(self):
#         df = self.spark.createDataFrame([{"a": {"b": "hello"}}, {"a": {"b": "world"}}], self.schema)
#         df = (
#             Replace(
#                 df=df,
#                 params={
#                     "source_column": "a.b",
#                     "original_value": "world",
#                     "new_value": "programmer",
#                 },
#                 target_alias="a.c",
#                 columns_to_flatten=["a{}.b"],
#             )
#             .execute()
#             .df
#         )
#         result = df.rdd.map(lambda r: r.a.c).collect()
#         self.assertEqual(sorted(result), ["hello", "programmer"])
#



"""
Teradata reader.
"""

from typing import Any, Dict, Optional

from koheesio.models import Field
from koheesio.spark.readers.jdbc import JdbcReader


class TeradataReader(JdbcReader):
    """
    Wrapper around JdbcReader for Teradata.

    Notes
    -----
    * Consider using synthetic partitioning column when using partitioned read:
        `MOD(HASHBUCKET(HASHROW(<TABLE>.<COLUMN>)), <NUM_PARTITIONS>)`
    * Relevant jars should be added to the Spark session manually. This class does not take care of that.

    See Also
    --------
    * Refer to [JdbcReader](jdbc.md#koheesio.spark.readers.jdbc.JdbcReader) for the list of all available
        parameters.
    * Refer to Teradata docs for the list of all available connection string parameters:
        https://teradata-docs.s3.amazonaws.com/doc/connectivity/jdbc/reference/current/jdbcug_chapter_2.html#BABJIHBJ

    Example
    -------
    This example depends on the Teradata `terajdbc4` JAR. e.g. terajdbc4-17.20.00.15. Keep in mind that older versions
    of `terajdbc4` drivers also require `tdgssconfig` JAR.

    ```python
    from koheesio.spark.readers.teradata import TeradataReader

    td = TeradataReader(
        url="jdbc:teradata://<domain_or_ip>/logmech=ldap,charset=utf8,database=<db>,type=fastexport, maybenull=on",
        user="YOUR_USERNAME",
        password="***",
        dbtable="schemaname.tablename",
    )
    ```

    Parameters
    ----------
    url : str
        JDBC connection string. Refer to Teradata docs for the list of all available connection string parameters.
        Example: `jdbc:teradata://<domain_or_ip>/logmech=ldap,charset=utf8,database=<db>,type=fastexport, maybenull=on`
    user : str
        Username
    password : SecretStr
        Password
    dbtable : str
        Database table name, also include schema name
    options : Optional[Dict[str, Any]], optional, default={"fetchsize": 2000, "numPartitions": 10}
        Extra options to pass to the Teradata JDBC driver. Refer to Teradata docs for the list of all available
        connection string parameters.
    query : Optional[str], optional, default=None
        Query
    format : str
        The type of format to load. Defaults to 'jdbc'. Should not be changed.
    driver : str
        Driver name. Be aware that the driver jar needs to be passed to the task. Should not be changed.
    """

    driver: str = Field(
        "com.teradata.jdbc.TeraDriver",
        description="Make sure that the necessary JARs are available in the cluster: terajdbc4-x.x.x.x",
    )
    options: Optional[Dict[str, Any]] = Field(
        {"fetchsize": 2000, "numPartitions": 10}, description="Extra options to pass to the Teradata JDBC driver"
    )



"""SQL Transform module

SQL Transform module provides an easy interface to transform a dataframe using SQL. This SQL can originate from a
string or a file and may contain placeholders for templating.
"""

from koheesio.models.sql import SqlBaseStep
from koheesio.spark.transformations import Transformation
from koheesio.utils import get_random_string


class SqlTransform(SqlBaseStep, Transformation):
    """
    SQL Transform module provides an easy interface to transform a dataframe using SQL.

    This SQL can originate from a string or a file and may contain placeholder (parameters) for templating.

    - Placeholders are identified with `${placeholder}`.
    - Placeholders can be passed as explicit params (params) or as implicit params (kwargs).

    Example sql script:

    ```sql
    SELECT id, id + 1 AS incremented_id, ${dynamic_column} AS extra_column
    FROM ${table_name}
    ```
    """

    def execute(self):
        table_name = get_random_string(prefix="sql_transform")
        self.params = {**self.params, "table_name": table_name}

        df = self.df
        df.createOrReplaceTempView(table_name)
        query = self.query

        self.output.df = self.spark.sql(query)



import pytest

from koheesio.logger import LoggingFactory
from koheesio.spark.transformations.uuid5 import HashUUID5

pytestmark = pytest.mark.spark

log = LoggingFactory.get_logger(name="test_uuid5")


@pytest.mark.parametrize(
    "input_values,expected",
    [
        (
            # description: base test
            # input values,
            dict(target_column="uuid5_hash", source_columns=["id", "string"]),
            # expected data
            [
                dict(id=1, string="hello", uuid5_hash="66bee5a3-651f-57fd-b070-b873c39ffb56"),
                dict(id=2, string="world", uuid5_hash="4a90c3a5-7173-5764-8905-94455c661e6c"),
                dict(id=3, string="", uuid5_hash="3ee49156-19d1-5335-b31f-8e11e9a17ed1"),
            ],
        ),
        (
            # description: test with custom namespace
            # input values,
            dict(target_column="uuid5_hash", source_columns=["id", "string"], namespace="unittest"),
            # expected data
            [
                dict(id=1, string="hello", uuid5_hash="fd12783f-ac26-5624-bbdd-dc0ae289d099"),
                dict(id=2, string="world", uuid5_hash="0e6862a7-6327-5277-9666-9adbb5dbd918"),
                dict(id=3, string="", uuid5_hash="741e42f0-baf6-547e-9f6d-cbf42dfb41e2"),
            ],
        ),
        (
            # description: test with extra string and namespace
            # input values,
            dict(target_column="id", source_columns=["id", "string"], namespace="unittest", extra_string="llama_drama"),
            # expected data
            [
                dict(id="f3e99bbd-85ae-5dc3-bf6e-cd0022a0ebe6", string="hello"),
                dict(id="b48e880f-c289-5c94-b51f-b9d21f9616c0", string="world"),
                dict(id="2193a99d-222e-5a0c-a7d6-48fbe78d2708", string=""),
            ],
        ),
        (
            # description: test with custom delimiter
            # input values,
            dict(target_column="string", source_columns=["id", "string"], delimiter=","),
            # expected data
            [
                dict(id=1, string="f0e5b058-78e0-5e33-b8a6-e434a0e9b8b0"),
                dict(id=2, string="5ad87a13-799c-5699-870e-10ce1196cc1b"),
                dict(id=3, string="81d87123-40fe-5612-b2c3-1c2e294ee636"),
            ],
        ),
    ],
)
def test_hash_uuid5(input_values, expected, sample_df_with_strings):
    df = HashUUID5(**input_values).transform(sample_df_with_strings)
    actual = [k.asDict() for k in df.collect()]
    assert actual == expected
    log.info("HashUUID5 unit test completed")



import pytest

from koheesio.models import ValidationError
from koheesio.spark.transformations.repartition import Repartition

pytestmark = pytest.mark.spark


@pytest.mark.parametrize(
    "input_values,expected",
    [
        (
            # case 1 : both column and num_partitions
            dict(column="country", num_partitions=2),
            # expected num_partitions
            2,
        ),
        (
            # case 2 : just column and no num_partitions
            dict(column="country"),
            # expected num_partitions
            1,
        ),
        (
            # case 3 : just num_partitions
            dict(num_partitions=3),
            # expected num_partitions
            3,
        ),
        (
            # case 4 : way too many num_partitions
            dict(num_partitions=5000),
            # expected num_partitions
            5000,
        ),
    ],
)
def test_repartition(input_values, expected, spark):
    input_df = spark.createDataFrame(
        [
            ["Banana", 1000, "USA"],
            ["Carrots", 1500, "USA"],
            ["Beans", 1600, "USA"],
            ["Orange", 1234, "USA"],
            ["Orange", 5555, "USA"],
            ["Banana", 400, "China"],
            ["Carrots", 1200, "China"],
            ["Beans", 1500, "China"],
            ["Orange", 4000, "China"],
            ["Banana", 2000, "Canada"],
            ["Carrots", 2000, "Canada"],
            ["Beans", 2000, "Mexico"],
        ],
        schema="product string, amount int, country string",
    )

    df = Repartition(**input_values).transform(input_df)
    assert df.rdd.getNumPartitions() == expected


def test_repartition_should_raise_error():
    with pytest.raises(ValidationError):
        Repartition()  # no input



import pytest

from koheesio.spark.transformations.get_item import GetItem

pytestmark = pytest.mark.spark


@pytest.mark.parametrize(
    "input_values,input_data,input_schema,expected",
    [
        (
            # Test GetItem with List
            # description: Test transform method when an integer/ordinal is passed
            #     Deliberately picked a number that is bigger than some of the lengths.
            #     Should return null and not fail.
            # input values
            dict(
                column="content",
                key=1,
                target_column="item",
            ),
            # input data
            [[1, [1, 2, 3]], [2, [4, 5]], [3, [6]], [4, []]],
            # input schema
            ["id", "content"],
            # expected data
            [2, 5, None, None],
        ),
        (
            # Test GetItem with Dict
            # description: Test transform method when a key/string is passed
            #     2 rows have key1 and 2 rows have keys as keys. Meaning
            #     that choosing one of the 2 will result the other 2 being null.
            # input values
            dict(
                column="content",
                key="key2",
                target_column="item",
            ),
            # input data
            [
                [1, {"key1": "value1"}],
                [2, {"key1": "value2"}],
                [3, {"key2": "hello"}],
                [4, {"key2": "world"}],
            ],
            # input schema
            ["id", "content"],
            # expected data
            [None, None, "hello", "world"],
        ),
    ],
)
def test_transform_get_item(input_values, input_data, input_schema, expected, spark):
    input_df = spark.createDataFrame(data=input_data, schema=input_schema)
    gi = GetItem(**input_values)
    output_df = gi.transform(input_df)
    actual = output_df.orderBy(input_schema[0]).select(gi.target_column).rdd.map(lambda r: r[0]).collect()
    assert actual == expected


# def test_transform_nested_df(self):
#     """
#     TODO: implement... this functionality is not supported yet
#     Check whether a nested dataframe can be accessed.
#     """
#     output_df = (
#         GetItem(
#             df=self.sample_nested_df,
#             params={
#                 "source_column": "payload.col2.array_col",
#                 "key_or_ordinal": 0,
#             },
#             target_alias="item",
#             columns_to_flatten=["payload{}.col2{}.array_col{}"],
#         )
#         .execute()
#         .df
#     )
#
#         output_list = output_df.select("item").rdd.map(lambda r: r[0]).collect()
#
#         self.assertEqual([Row(item="v1")], output_list)



import gzip
from datetime import datetime, timezone
from importlib.util import find_spec

import pytest

from pyspark.sql.types import (
    BooleanType,
    FloatType,
    IntegerType,
    StructField,
    StructType,
    TimestampType,
)

from koheesio.spark.writers.buffer import PandasCsvBufferWriter, PandasJsonBufferWriter

pytestmark = pytest.mark.spark

# Test data has two columns: email and sha256_email
test_data = [
    (
        "john.doe@email.com",
        "55f537baf75630a85fda6540cdf43859a51ab13ba33dc314600618f89a6ffa0e",
    ),
    (
        "jane.doe@email.com",
        "390852f40addcb9e2868f180b471ef4315a34063f145447d73f51f5e5ad1ef47",
    ),
    (
        "foo@bar.baz",
        "80c66bdd90ae7fd4378cef780422fe428ee7fb526301f7b236113c4ece3be146",
    ),
]
test_schema = ["email", "sha256_email"]


def test_canary():
    """Check that pandas is installed"""
    assert find_spec("pandas") is not None, "pandas is not installed"


@pytest.mark.parametrize(
    "data,schema,params,expected_data",
    [
        (
            test_data,
            test_schema,
            {"header": True},
            "email,sha256_email\n"
            "john.doe@email.com,55f537baf75630a85fda6540cdf43859a51ab13ba33dc314600618f89a6ffa0e\n"
            "jane.doe@email.com,390852f40addcb9e2868f180b471ef4315a34063f145447d73f51f5e5ad1ef47\n"
            "foo@bar.baz,80c66bdd90ae7fd4378cef780422fe428ee7fb526301f7b236113c4ece3be146",
        ),
        (
            # Change the separator to ; and no header
            test_data,
            test_schema,
            {"header": False, "sep": ";"},
            "john.doe@email.com;55f537baf75630a85fda6540cdf43859a51ab13ba33dc314600618f89a6ffa0e\n"
            "jane.doe@email.com;390852f40addcb9e2868f180b471ef4315a34063f145447d73f51f5e5ad1ef47\n"
            "foo@bar.baz;80c66bdd90ae7fd4378cef780422fe428ee7fb526301f7b236113c4ece3be146",
        ),
        (
            # test index, and columns
            test_data,
            test_schema,
            {"header": True, "index": True, "index_label": "i", "columns": ["email"]},
            "i,email\n" "0,john.doe@email.com\n" "1,jane.doe@email.com\n" "2,foo@bar.baz",
        ),
        (
            # test with compression
            test_data,
            test_schema,
            {"header": True, "compression": "gzip"},
            "email,sha256_email\n"
            "john.doe@email.com,55f537baf75630a85fda6540cdf43859a51ab13ba33dc314600618f89a6ffa0e\n"
            "jane.doe@email.com,390852f40addcb9e2868f180b471ef4315a34063f145447d73f51f5e5ad1ef47\n"
            "foo@bar.baz,80c66bdd90ae7fd4378cef780422fe428ee7fb526301f7b236113c4ece3be146",
        ),
        (
            # test with timestamp_format
            [(datetime(1964, 1, 25, tzinfo=timezone.utc),), (datetime(1971, 5, 30, tzinfo=timezone.utc),)],
            StructType([StructField("date", TimestampType())]),
            {"header": False},  # default timestamp_format is "%Y-%m-%dT%H:%M:%S.%f"
            "1964-01-25T00:00:00.000000\n1971-05-30T00:00:00.000000",
        ),
    ],
)
def test_pandas_csv_buffer_writer(data, schema, params, expected_data, spark):
    writer = PandasCsvBufferWriter(**params)

    # Execute the writer
    writer.write(
        df=spark.createDataFrame(
            data,
            schema,
        )
    )
    # Get the CSV string from the buffer
    actual = writer.output.read()

    if params.get("compression"):
        # Check that the compression is correct by checking the first two bytes (magic number)
        assert actual.startswith(b"\x1f\x8b")
        # Decompress the data to check the content
        actual = gzip.decompress(actual)

    # Check that the CSV string matches the expected data
    assert actual.decode("utf-8").strip() == expected_data


@pytest.mark.parametrize(
    "data,schema,params,expected_data",
    [
        (
            # Test with default parameters
            test_data,
            test_schema,
            {},  # should be the same as {"orient": "records", "lines": True},
            '{"email":"john.doe@email.com",'
            '"sha256_email":"55f537baf75630a85fda6540cdf43859a51ab13ba33dc314600618f89a6ffa0e"}\n'
            '{"email":"jane.doe@email.com",'
            '"sha256_email":"390852f40addcb9e2868f180b471ef4315a34063f145447d73f51f5e5ad1ef47"}\n'
            '{"email":"foo@bar.baz",'
            '"sha256_email":"80c66bdd90ae7fd4378cef780422fe428ee7fb526301f7b236113c4ece3be146"}',
        ),
        (
            # Test with orient="split", only one column
            test_data,
            test_schema,
            {"orient": "split", "columns": ["email"]},
            "{"
            '"columns":["email"],'
            '"index":[0,1,2],'
            '"data":[["john.doe@email.com"],["jane.doe@email.com"],["foo@bar.baz"]]'
            "}",
        ),
        (
            # Test with orient="index"
            test_data,
            test_schema,
            {"orient": "index"},
            "{"
            '"0":{'
            '"email":"john.doe@email.com",'
            '"sha256_email":"55f537baf75630a85fda6540cdf43859a51ab13ba33dc314600618f89a6ffa0e"},'
            '"1":{'
            '"email":"jane.doe@email.com",'
            '"sha256_email":"390852f40addcb9e2868f180b471ef4315a34063f145447d73f51f5e5ad1ef47"},'
            '"2":{'
            '"email":"foo@bar.baz",'
            '"sha256_email":"80c66bdd90ae7fd4378cef780422fe428ee7fb526301f7b236113c4ece3be146"}'
            "}",
        ),
        (
            # Test with orient="columns", only one column
            test_data,
            test_schema,
            {"orient": "columns", "columns": ["email"]},
            '{"email":{"0":"john.doe@email.com","1":"jane.doe@email.com","2":"foo@bar.baz"}}',
        ),
        (
            # Test with orient="values"
            test_data,
            test_schema,
            {"orient": "values"},
            "["
            '["john.doe@email.com","55f537baf75630a85fda6540cdf43859a51ab13ba33dc314600618f89a6ffa0e"],'
            '["jane.doe@email.com","390852f40addcb9e2868f180b471ef4315a34063f145447d73f51f5e5ad1ef47"],'
            '["foo@bar.baz","80c66bdd90ae7fd4378cef780422fe428ee7fb526301f7b236113c4ece3be146"]'
            "]",
        ),
        (
            # Test with orient="records", and lines=False, only one column
            test_data,
            test_schema,
            {"orient": "records", "lines": False, "columns": ["email"]},
            '[{"email":"john.doe@email.com"},{"email":"jane.doe@email.com"},{"email":"foo@bar.baz"}]',
        ),
        (
            # Test with orient="table"
            test_data,
            test_schema,
            {"orient": "table"},
            "{"
            '"schema":{'
            '"fields":[{"name":"index","type":"integer"},'
            '{"name":"email","type":"string"},'
            '{"name":"sha256_email","type":"string"}],'
            '"primaryKey":["index"],'
            '"pandas_version":"1.4.0"},'
            '"data":['
            '{"index":0,"email":"john.doe@email.com",'
            '"sha256_email":"55f537baf75630a85fda6540cdf43859a51ab13ba33dc314600618f89a6ffa0e"},'
            '{"index":1,"email":"jane.doe@email.com",'
            '"sha256_email":"390852f40addcb9e2868f180b471ef4315a34063f145447d73f51f5e5ad1ef47"},'
            '{"index":2,"email":"foo@bar.baz",'
            '"sha256_email":"80c66bdd90ae7fd4378cef780422fe428ee7fb526301f7b236113c4ece3be146"}]}',
        ),
        (
            # Test with integer, float, and boolean columns
            [
                {"age": 30, "rating": 4.5, "is_active": True},
                {"age": 25, "rating": 4.7, "is_active": False},
                {"age": 35, "rating": 4.3, "is_active": True},
            ],
            StructType(
                [
                    StructField("age", IntegerType(), True),
                    StructField("rating", FloatType(), True),
                    StructField("is_active", BooleanType(), True),
                ]
            ),
            {"double_precision": 2},
            '{"age":30,"rating":4.5,"is_active":true}\n'
            '{"age":25,"rating":4.7,"is_active":false}\n'
            '{"age":35,"rating":4.3,"is_active":true}',
        ),
    ],
)
def test_pandas_json_buffer_writer(data, schema, params, expected_data, spark):
    writer = PandasJsonBufferWriter(**params)

    # Execute the writer
    writer.write(
        df=spark.createDataFrame(
            data,
            schema,
        )
    )

    # Get the JSON string from the buffer
    json_str = writer.output.read().decode("utf-8")

    # Check that the JSON string matches the expected data
    assert json_str.strip() == expected_data


def test_pandas_json_with_compression(spark):
    writer = PandasJsonBufferWriter(compression="gzip")

    # Execute the writer
    writer.write(
        df=spark.createDataFrame(
            test_data,
            test_schema,
        )
    )

    # Check that the compression is correct by checking the first two bytes (magic number)
    assert writer.output.read().startswith(b"\x1f\x8b")

    # Decompress the data to check the content
    json_str = gzip.decompress(writer.output.read()).decode("utf-8")

    # Check that the JSON string matches the expected data
    assert json_str.strip() == (
        '{"email":"john.doe@email.com",'
        '"sha256_email":"55f537baf75630a85fda6540cdf43859a51ab13ba33dc314600618f89a6ffa0e"}\n'
        '{"email":"jane.doe@email.com",'
        '"sha256_email":"390852f40addcb9e2868f180b471ef4315a34063f145447d73f51f5e5ad1ef47"}\n'
        '{"email":"foo@bar.baz",'
        '"sha256_email":"80c66bdd90ae7fd4378cef780422fe428ee7fb526301f7b236113c4ece3be146"}'
    )



"""Steps Module

This module contains the definition of the `Step` class, which serves as the base class for custom units of logic that
can be executed. It also includes the `StepOutput` class, which defines the output data model for a `Step`.

The `Step` class is designed to be subclassed for creating new steps in a data pipeline. Each subclass should implement
the `execute` method, specifying the expected inputs and outputs.

This module also exports the `SparkStep` class for steps that interact with Spark

Classes:
--------
- *Step*: Base class for a custom unit of logic that can be executed.
- *StepOutput*: Defines the output data model for a `Step`.
"""

from __future__ import annotations

import inspect
import json
import sys
import warnings
from typing import Any
from abc import abstractmethod
from functools import partialmethod, wraps

import yaml

from koheesio.models import BaseModel, ConfigDict, ModelMetaclass

__all__ = [
    "Step",
    "StepMetaClass",
    "StepOutput",
]


class StepOutput(BaseModel):
    """Class for the StepOutput model

    Usage
    -----
    Setting up the StepOutputs class is done like this:
    ```python
    class YourOwnOutput(StepOutput):
        a: str
        b: int
    ```
    """

    model_config = ConfigDict(
        validate_default=False,
        defer_build=True,
    )

    # pylint: disable=no-value-for-parameter
    def validate_output(self) -> StepOutput:
        """Validate the output of the Step

        Essentially, this method is a wrapper around the validate method of the BaseModel class
        """
        validated_model = self.validate()
        return StepOutput.from_basemodel(validated_model)


class StepMetaClass(ModelMetaclass):
    """
    StepMetaClass has to be set up as a Metaclass extending ModelMetaclass to allow Pydantic to be unaffected while
    allowing for the execute method to be auto-decorated with do_execute
    """

    # Solution to overcome issue with python>=3.11,
    # When partialmethod is forgetting that _execute_wrapper
    # is a method of wrapper, and it needs to pass that in as the first arg.
    # https://github.com/python/cpython/issues/99152
    class _partialmethod_with_self(partialmethod):
        def __get__(self, obj, cls=None):
            return self._make_unbound_method().__get__(obj, cls)

    # Unique object to mark a function as wrapped
    _step_execute_wrapper_sentinel = object()

    # pylint: disable=signature-differs
    def __new__(
        mcs,
        cls_name: str,
        bases: tuple[type[Any], ...],
        namespace: dict[str, Any],
        **kwargs: Any,
    ) -> type:
        """
        Create a new class instance.

        Parameters
        ----------
        mcs : type
            The metaclass of the class being created.
        cls_name : str
            The name of the class being created.
        bases : tuple[type[Any], ...]
            The base classes of the class being created.
        namespace : dict[str, Any]
            The namespace of the class being created.
        kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        type
            The newly created class instance.

        Notes
        -----
        This method is called before the class is created. It allows customization of the class creation process.

        The method wraps the `execute` method of the class with a partial method if it is not already wrapped.
        The wrapped method is then set as the new `execute` method of the class.

        If the `execute` method is already wrapped, the method is not modified.

        The method also keeps track of the number of times the `execute` method has been wrapped.

        """
        cls = super().__new__(
            mcs,
            cls_name,
            bases,
            namespace,
            **kwargs,
        )

        # Extract execute method present in the class
        execute_method = getattr(cls, "execute")

        # check if function is already wrapped with do_execute
        # Here we are trying to get the attribute "_partialmethod__step_execute_wrapper_sentinel"
        # from the execute_method function.
        # If the function is already wrapped, this attribute should exist.
        sentinel = getattr(execute_method, "_partialmethod__step_execute_wrapper_sentinel", None)

        # Check if the sentinel is the same as the class's sentinel. If they are the same,
        # it means the function is already wrapped.
        is_already_wrapped = sentinel is cls._step_execute_wrapper_sentinel

        # Get the wrap count of the function. If the function is not wrapped yet, the default value is 0.
        wrap_count = getattr(execute_method, "_partialmethod_wrap_count", 0)

        # prevent multiple wrapping
        # If the function is not already wrapped, we proceed to wrap it.
        if not is_already_wrapped:
            # Create a partial method with the execute_method as one of the arguments.
            # This is the new function that will be called instead of the original execute_method.
            wrapper = mcs._partialmethod_impl(cls=cls, execute_method=execute_method)

            # Updating the attributes of the wrapping function to those of the original function.
            wraps(execute_method)(wrapper)  # type: ignore

            # Set the sentinel attribute to the wrapper. This is done so that we can check
            # if the function is already wrapped.
            setattr(wrapper, "_step_execute_wrapper_sentinel", cls._step_execute_wrapper_sentinel)

            # Increase the wrap count of the function. This is done to keep track of
            # how many times the function has been wrapped.
            setattr(wrapper, "_wrap_count", wrap_count + 1)
            setattr(cls, "execute", wrapper)

        return cls

    @staticmethod
    def _is_called_through_super(caller_self: Any, caller_name: str, *_args, **_kwargs) -> bool:
        """
        Check if the method is called through super() in the immediate parent class.

        Parameters
        ----------
        caller_self : Any
            The instance of the class.
        caller_name : str
            The name of the caller method.
        *args : Any
            Additional positional arguments.
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        bool
            True if the method is called through super(), False otherwise.

        """

        base_class = caller_self.__class__.__bases__[0]
        return caller_name in base_class.__dict__

    @classmethod
    def _partialmethod_impl(mcs, cls: type, execute_method) -> partialmethod:
        """
        This method creates a partial method implementation for a given class and execute method.
        It handles a specific issue with python>=3.11 where partialmethod forgets that _execute_wrapper
        is a method of wrapper, and it needs to pass that in as the first argument.

        Args:
            mcs: The metaclass instance.
            cls (type): The class for which the partial method is being created.
            execute_method: The method to be executed.

        Returns:
            wrapper: The partial method implementation.
        """

        # Solution to overcome issue with python>=3.11,
        # When partialmethod is forgetting that _execute_wrapper
        # is a method of wrapper, and it needs to pass that in as the first arg.
        # https://github.com/python/cpython/issues/99152
        class _partialmethod_with_self(partialmethod):
            """
            This class is a workaround for the issue with python>=3.11 where partialmethod forgets that
            _execute_wrapper is a method of wrapper, and it needs to pass that in as the first argument.
            """

            def __get__(self, obj, cls=None):
                """
                This method returns the unbound method for the given object and class.

                Args:
                    obj: The object instance.
                    cls (Optional): The class for which the method is being retrieved. Defaults to None.

                Returns:
                    The unbound method.
                """
                return self._make_unbound_method().__get__(obj, cls)

        _partialmethod_impl = partialmethod if sys.version_info < (3, 11) else _partialmethod_with_self
        wrapper = _partialmethod_impl(cls._execute_wrapper, execute_method=execute_method)

        return wrapper

    @classmethod
    def _execute_wrapper(mcs, step: Step, execute_method, *args, **kwargs) -> StepOutput:
        """
        Method that wraps some common functionalities on Steps
        Ensures proper logging and makes it so that a Steps execute method always returns the StepOutput

        Parameters
        ----------
        step : Step
            The step instance.
        execute_method : Callable
            The execute method of the step.
        *args : Any
            Additional positional arguments.
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        StepOutput
            The output of the step.

        """

        # check if the method is called through super() in the immediate parent class
        caller_name = inspect.currentframe().f_back.f_back.f_code.co_name
        is_called_through_super_ = mcs._is_called_through_super(step, caller_name)

        mcs._log_start_message(step=step, skip_logging=is_called_through_super_)
        return_value = mcs._run_execute(step=step, execute_method=execute_method, *args, **kwargs)
        mcs._configure_step_output(step=step, return_value=return_value)
        mcs._validate_output(step=step, skip_validating=is_called_through_super_)
        mcs._log_end_message(step=step, skip_logging=is_called_through_super_)

        return step.output

    @classmethod
    def _log_start_message(mcs, step: Step, *_args, skip_logging: bool = False, **_kwargs):
        """
        Log the start message of the step execution

        Parameters
        ----------
        step : Step
            The step instance.
        skip_logging : bool, optional
            Whether to skip logging, by default False.
        *args : Any
            Additional positional arguments.
        **kwargs : Any
            Additional keyword arguments.

        """

        if not skip_logging:
            step.log.info("Start running step")
            step.log.debug(f"Step Input: {step.__repr_str__(' ')}")

    @classmethod
    def _log_end_message(mcs, step: Step, *_args, skip_logging: bool = False, **_kwargs):
        """
        Log the end message of the step execution

        Parameters
        ----------
        step : Step
            The step instance.
        skip_logging : bool, optional
            Whether to skip logging, by default False.
        *args : Any
            Additional positional arguments.
        **kwargs : Any
            Additional keyword arguments.

        """

        if not skip_logging:
            step.log.debug(f"Step Output: {step.output.__repr_str__(' ')}")
            step.log.info("Finished running step")

    @classmethod
    def _validate_output(mcs, step: Step, *_args, skip_validating: bool = False, **_kwargs):
        """
        Validate the output of the step

        Parameters
        ----------
        step : Step
            The step instance.
        skip_validating : bool, optional
            Whether to skip validating, by default False.
        *args : Any
            Additional positional arguments.
        **kwargs : Any
            Additional keyword arguments.

        """

        if not skip_validating:
            step.output.validate_output()

    @classmethod
    def _configure_step_output(mcs, step, return_value: Any, *_args, **_kwargs):
        """
        Configure the output of the step.
        If the execute method returns a value, and it is not the output, set the output to the return value

        Parameters
        ----------
        step : Step
            The step instance.
        return_value : Any
            The return value of the execute method.
        *args : Any
            Additional positional arguments.
        **kwargs : Any
            Additional keyword arguments.

        """

        output: StepOutput = step.output

        if return_value:
            if not isinstance(return_value, StepOutput):
                msg = (
                    f"execute() did not produce output of type {output.name}, returns of the wrong type will be ignored"
                )
                warnings.warn(msg)
                step.log.warning(msg)

            if return_value != output:
                step.output.merge(return_value)

        step.output = output

    @classmethod
    def _run_execute(mcs, execute_method, step, *args, **kwargs) -> Any:
        """
        Run the execute method of the step, and catch any errors

        Parameters
        ----------
        execute_method : Callable
            The execute method of the step.
        step : Step
            The step instance.
        *args : Any
            Additional positional arguments.
        **kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        Any
            The return value of the execute method.

        """

        try:
            return_value = execute_method(step, *args, **kwargs)
        except Exception as e:
            step.log.error(f"Error while running step: \n{step.__repr_str__(join_str=' ')}")
            raise e

        return return_value


class Step(BaseModel, metaclass=StepMetaClass):
    """Base class for a step

    A custom unit of logic that can be executed.

    The Step class is designed to be subclassed. To create a new step, one would subclass Step and implement the
    `def execute(self)` method, specifying the expected inputs and outputs.

    > Note: since the Step class is meta classed, the execute method is wrapped with the `do_execute` function making
    > it always return the Step's output. Hence, an explicit return is not needed when implementing execute.

    Methods and Attributes
    ----------------------
    The Step class has several attributes and methods.

    #### INPUT
    The following fields are available by default on the Step class:
    - `name`: Name of the Step. If not set, the name of the class will be used.
    - `description`: Description of the Step. If not set, the docstring of the class will be used. If the docstring
        contains multiple lines, only the first line will be used.

    When subclassing a Step, any additional pydantic field will be treated as `input` to the Step. See also the
    explanation on the `.execute()` method below.


    #### OUTPUT
    Every Step has an `Output` class, which is a subclass of `StepOutput`. This class is used to validate the output of
    the Step. The `Output` class is defined as an inner class of the Step class. The `Output` class can be accessed
    through the `Step.Output` attribute. The `Output` class can be extended to add additional fields to the output of
    the Step. See also the explanation on the `.execute()`.

    - `Output`: A nested class representing the output of the Step used to validate the output of the
        Step and based on the StepOutput class.
    - `output`: Allows you to interact with the Output of the Step lazily (see above and StepOutput)

    When subclassing a Step, any additional pydantic field added to the nested `Output` class will be treated as
    `output` of the Step. See also the description of `StepOutput` for more information.


    #### Methods:
    - `execute`: Abstract method to implement for new steps.
        - The Inputs of the step can be accessed, using `self.input_name`.
        - The output of the step can be accessed, using `self.output.output_name`.
    - `run`: Alias to .execute() method. You can use this to run the step, but execute is preferred.
    - `to_yaml`: YAML dump the step
    - `get_description`: Get the description of the Step

    When subclassing a Step, `execute` is the only method that __needs__ to be implemented. Any additional method added
    to the class will be treated as a method of the Step.

    Note: since the Step class is meta-classed, the execute method is automatically wrapped with the `do_execute`
    function making it always return a StepOutput. See also the explanation on the `do_execute` function.


    #### class methods:
    - `from_step`: Returns a new Step instance based on the data of another Step instance.
        for example: `MyStep.from_step(other_step, a="foo")`
    - `get_description`: Get the description of the Step


    #### dunder methods:
    - `__getattr__`: Allows input to be accessed through `self.input_name`
    - `__repr__` and `__str__`: String representation of a step


    Background
    ----------
    A Step is an atomic operation and serves as the building block of data pipelines built with the framework.
    Tasks typically consist of a series of Steps.

    A step can be seen as an operation on a set of inputs, that returns a set of outputs. This however does not imply
    that steps are stateless (e.g. data writes)!

    The diagram serves to illustrate the concept of a Step:

    ```text
    ┌─────────┐        ┌──────────────────┐        ┌─────────┐
    │ Input 1 │───────▶│                  ├───────▶│Output 1 │
    └─────────┘        │                  │        └─────────┘
                       │                  │
    ┌─────────┐        │                  │        ┌─────────┐
    │ Input 2 │───────▶│       Step       │───────▶│Output 2 │
    └─────────┘        │                  │        └─────────┘
                       │                  │
    ┌─────────┐        │                  │        ┌─────────┐
    │ Input 3 │───────▶│                  ├───────▶│Output 3 │
    └─────────┘        └──────────────────┘        └─────────┘
    ```

    Steps are built on top of Pydantic, which is a data validation and settings management using python type
    annotations. This allows for the automatic validation of the inputs and outputs of a Step.

    - Step inherits from BaseModel, which is a Pydantic class used to define data models. This allows Step to
      automatically validate data against the defined fields and their types.
    - Step is metaclassed by StepMetaClass, which is a custom metaclass that wraps the `execute` method of the Step
      class with the `_execute_wrapper` function. This ensures that the `execute` method always returns the output of
      the Step along with providing logging and validation of the output.
    - Step has an `Output` class, which is a subclass of `StepOutput`. This class is used to validate the output
      of the Step. The `Output` class is defined as an inner class of the Step class. The `Output` class can be
      accessed through the `Step.Output` attribute.
    - The `Output` class can be extended to add additional fields to the output of the Step.

    Examples
    --------
    ```python
    class MyStep(Step):
        a: str  # input

        class Output(StepOutput):  # output
            b: str

        def execute(self) -> MyStep.Output:
            self.output.b = f"{self.a}-some-suffix"
    ```
    """

    class Output(StepOutput):
        """Output class for Step"""

    __output__: Output

    @property
    def output(self) -> Output:
        """Interact with the output of the Step"""
        if not self.__output__:
            self.__output__ = self.Output.lazy()
            self.__output__.name = self.name + ".Output"
            self.__output__.description = "Output for " + self.name
        return self.__output__

    @output.setter
    def output(self, value: Output):
        """Set the output of the Step"""
        self.__output__ = value

    @abstractmethod
    def execute(self):
        """Abstract method to implement for new steps.

        The Inputs of the step can be accessed, using `self.input_name`

        Note: since the Step class is meta-classed, the execute method is wrapped with the `do_execute` function making
          it always return the Steps output
        """
        raise NotImplementedError

    def run(self):
        """Alias to .execute()"""
        return self.execute()

    def __repr__(self) -> str:
        """String representation of a step"""
        class_name = self.__class__.__name__
        underline = "=" * len(class_name)
        yaml_output = self.repr_yaml(simple=True)
        return f"{class_name}\n{underline}\n{yaml_output}"

    def __str__(self) -> str:
        """String representation of a step"""
        return self.__repr__()

    def repr_json(self, simple=False) -> str:
        """dump the step to json, meant for representation

        Note: use to_json if you want to dump the step to json for serialization
        This method is meant for representation purposes only!

        Examples
        --------
        ```python
        >>> step = MyStep(a="foo")
        >>> print(step.repr_json())
        {"input": {"a": "foo"}}
        ```

        Parameters
        ----------
        simple: bool
            When toggled to True, a briefer output will be produced. This is friendlier for logging purposes

        Returns
        -------
        str
            A string, which is valid json
        """
        model_dump_options = dict(warnings="none", exclude_unset=True)

        _result = {}

        # extract input
        _input = self.model_dump(**model_dump_options)

        # remove name and description from input and add to result if simple is not set
        name = _input.pop("name", None)
        description = _input.pop("description", None)
        if not simple:
            if name:
                _result["name"] = name
            if description:
                _result["description"] = description
        else:
            model_dump_options["exclude"] = {"name", "description"}

        # extract output
        _output = self.output.model_dump(**model_dump_options)

        # add output to result
        if _output:
            _result["output"] = _output

        # add input to result
        _result["input"] = _input

        class MyEncoder(json.JSONEncoder):
            """Custom JSON Encoder to handle non-serializable types"""

            def default(self, o: Any) -> Any:
                try:
                    return super().default(o)
                except TypeError:
                    return o.__class__.__name__

        # Use MyEncoder when converting the dictionary to a JSON string
        json_str = json.dumps(_result, cls=MyEncoder)

        return json_str

    def repr_yaml(self, simple=False) -> str:
        """dump the step to yaml, meant for representation

        Note: use to_yaml if you want to dump the step to yaml for serialization
        This method is meant for representation purposes only!

        Examples
        --------
        ```python
        >>> step = MyStep(a="foo")
        >>> print(step.repr_yaml())
        input:
          a: foo
        ```

        Parameters
        ----------
        simple: bool
            When toggled to True, a briefer output will be produced. This is friendlier for logging purposes

        Returns
        -------
        str
            A string, which is valid yaml
        """
        json_str = self.repr_json(simple=simple)

        # Parse the JSON string back into a dictionary
        _result = json.loads(json_str)

        return yaml.dump(_result)

    def __getattr__(self, key: str):
        """__getattr__ dunder

        Allows input to be accessed through `self.input_name`

        Parameters
        ----------
        key: str
            Name of the attribute to return the value of

        Returns
        -------
        Any
            The value of the attribute
        """
        return self.model_dump().get(key)

    @classmethod
    def from_step(cls, step: Step, **kwargs):
        """Returns a new Step instance based on the data of another Step or BaseModel instance"""
        return cls.from_basemodel(step, **kwargs)



from typing import Any, Dict

import pytest

from pyspark.sql import DataFrame
from pyspark.sql import functions as f

from koheesio.logger import LoggingFactory
from koheesio.spark.transformations.transform import Transform

pytestmark = pytest.mark.spark

log = LoggingFactory.get_logger(name="test_transform")


def dummy_transform_func(df: DataFrame, target_column: str, value: str):
    return df.withColumn(target_column, f.lit(value))


def no_kwargs_dummy_func(df: DataFrame):
    return df


def transform_output_test(sdf: DataFrame, expected_data: Dict[str, Any]):
    return sdf.head().asDict() == expected_data


def test_dummy_function(dummy_df):
    # testing the dummy functions unwrapped / outside Transform class
    df = dummy_transform_func(dummy_df, "hello", "world")
    assert transform_output_test(df, {"id": 0, "hello": "world"})

    df = no_kwargs_dummy_func(dummy_df)
    assert transform_output_test(df, {"id": 0})


def test_verbose_transform(dummy_df):
    # verbose style input in Transform
    log.info(f"dummy_df: {dummy_df}")
    df = Transform(df=dummy_df, func=dummy_transform_func, params={"target_column": "foo", "value": "bar"}).execute().df
    assert transform_output_test(df, {"id": 0, "foo": "bar"})


def test_short_notation_on_transform(dummy_df):
    # shortened style notation (easier to read)
    df = Transform(df=dummy_df, func=dummy_transform_func, target_column="llama", value="drama").execute().df
    assert transform_output_test(df, {"id": 0, "llama": "drama"})


def test_ignore_too_much_input(dummy_df):
    # when too much input is given, transform should ignore extra input
    df = Transform(
        dummy_transform_func,
        target_column="so long",
        # ignored input
        value="and thanks for all the fish",
        title=42,
        author="Adams",
    ).transform(dummy_df)
    assert transform_output_test(df, {"id": 0, "so long": "and thanks for all the fish"})


def test_order_of_params(dummy_df):
    # order of params input should not matter
    df = Transform(
        dummy_transform_func,
        # out of position and extraneous
        value="lorem",
        title="thing",
        weekday="Wednesday",
        target_column="ipsum",
    ).transform(dummy_df)
    assert transform_output_test(df, {"id": 0, "ipsum": "lorem"})


def test_no_kwargs_function(dummy_df):
    df = Transform(no_kwargs_dummy_func).transform(dummy_df)
    assert transform_output_test(df, {"id": 0})


def test_from_func(dummy_df):
    # noinspection PyPep8Naming
    AddFooColumn = Transform.from_func(dummy_transform_func, target_column="foo")
    df = AddFooColumn(value="bar").transform(dummy_df)
    assert transform_output_test(df, {"id": 0, "foo": "bar"})



from datetime import datetime
from pathlib import Path, PurePath

import pytest

from koheesio.integrations.spark.tableau.hyper import (
    NOT_NULLABLE,
    NULLABLE,
    HyperFileDataFrameWriter,
    HyperFileListWriter,
    HyperFileParquetWriter,
    HyperFileReader,
    SqlType,
    TableDefinition,
    TableName,
)

pytestmark = pytest.mark.spark


class TestHyper:
    @pytest.fixture()
    def parquet_file(self, data_path):
        path = f"{data_path}/readers/parquet_file"
        return Path(path).glob("**/*.parquet")

    @pytest.fixture()
    def hyper_file(self, data_path):
        return f"{data_path}/readers/hyper_file/dummy.hyper"

    def test_hyper_file_reader(self, hyper_file):
        df = (
            HyperFileReader(
                path=hyper_file,
            )
            .execute()
            .df
        )

        assert df.count() == 3
        assert df.dtypes == [("string", "string"), ("int", "int"), ("timestamp", "timestamp")]

    def test_hyper_file_list_writer(self, spark):
        hw = HyperFileListWriter(
            name="test",
            table_definition=TableDefinition(
                table_name=TableName("Extract", "Extract"),
                columns=[
                    TableDefinition.Column(name="string", type=SqlType.text(), nullability=NOT_NULLABLE),
                    TableDefinition.Column(name="int", type=SqlType.int(), nullability=NULLABLE),
                    TableDefinition.Column(name="timestamp", type=SqlType.timestamp(), nullability=NULLABLE),
                ],
            ),
            data=[
                ["text_1", 1, datetime(2024, 1, 1, 0, 0, 0, 0)],
                ["text_2", 2, datetime(2024, 1, 2, 0, 0, 0, 0)],
                ["text_3", None, None],
            ],
        ).execute()

        df = (
            HyperFileReader(
                path=PurePath(hw.hyper_path),
            )
            .execute()
            .df
        )

        assert df.count() == 3
        assert df.dtypes == [("string", "string"), ("int", "int"), ("timestamp", "timestamp")]

    def test_hyper_file_parquet_writer(self, data_path, parquet_file):
        hw = HyperFileParquetWriter(
            name="test",
            table_definition=TableDefinition(
                table_name=TableName("Extract", "Extract"),
                columns=[
                    TableDefinition.Column(name="string", type=SqlType.text(), nullability=NOT_NULLABLE),
                    TableDefinition.Column(name="int", type=SqlType.int(), nullability=NULLABLE),
                    TableDefinition.Column(name="timestamp", type=SqlType.timestamp(), nullability=NULLABLE),
                ],
            ),
            files=parquet_file,
        ).execute()

        df = HyperFileReader(path=PurePath(hw.hyper_path)).execute().df

        assert df.count() == 6
        assert df.dtypes == [("string", "string"), ("int", "int"), ("timestamp", "timestamp")]

    def test_hyper_file_dataframe_writer(self, data_path, df_with_all_types):
        hw = HyperFileDataFrameWriter(
            name="test",
            df=df_with_all_types.drop("void", "byte", "binary", "array", "map", "float"),
        ).execute()

        df = HyperFileReader(path=PurePath(hw.hyper_path)).execute().df
        assert df.count() == 1
        assert df.dtypes == [
            ("short", "smallint"),
            ("integer", "int"),
            ("long", "bigint"),
            ("double", "float"),
            ("decimal", "decimal(18,5)"),
            ("string", "string"),
            ("boolean", "boolean"),
            ("timestamp", "timestamp"),
            ("date", "date"),
        ]



import os
import time
import uuid
from pathlib import Path

import pytest

from koheesio.logger import LoggingFactory
from koheesio.utils import get_project_root

if os.name != "nt":  # 'nt' is the name for Windows
    # force time zone to be UTC
    os.environ["TZ"] = "UTC"
    time.tzset()


PROJECT_ROOT = get_project_root()

TEST_DATA_PATH = Path(PROJECT_ROOT / "tests" / "_data")
DELTA_FILE = Path(TEST_DATA_PATH / "readers" / "delta_file")


@pytest.fixture(scope="session")
def random_uuid():
    return str(uuid.uuid4()).replace("-", "_")


@pytest.fixture(scope="session")
def logger(random_uuid):
    return LoggingFactory.get_logger(name="conf_test" + random_uuid)


@pytest.fixture(scope="session")
def data_path():
    return TEST_DATA_PATH.as_posix()


@pytest.fixture(scope="session")
def delta_file():
    return DELTA_FILE.as_posix()



"""
This module contains the SparkSqlReader class which reads the SparkSQL compliant query and returns the dataframe.
"""

from koheesio.models.sql import SqlBaseStep
from koheesio.spark.readers import Reader


class SparkSqlReader(SqlBaseStep, Reader):
    """
    SparkSqlReader reads the SparkSQL compliant query and returns the dataframe.

    This SQL can originate from a string or a file and may contain placeholder (parameters) for templating.
    - Placeholders are identified with ${placeholder}.
    - Placeholders can be passed as explicit params (params) or as implicit params (kwargs).

    Example
    -------
    SQL script (example.sql):
    ```sql
    SELECT id, id + 1 AS incremented_id, ${dynamic_column} AS extra_column
    FROM ${table_name}
    ```

    Python code:
    ```python
    from koheesio.spark.readers import SparkSqlReader

    reader = SparkSqlReader(
        sql_path="example.sql",
        # params can also be passed as kwargs
        dynamic_column"="name",
        "table_name"="my_table"
    )
    reader.execute()
    ```

    In this example, the SQL script is read from a file and the placeholders are replaced with the given params.
    The resulting SQL query is:
    ```sql
    SELECT id, id + 1 AS incremented_id, name AS extra_column
    FROM my_table
    ```

    The query is then executed and the resulting DataFrame is stored in the `output.df` attribute.

    Parameters
    ----------
    sql_path : str or Path
        Path to a SQL file
    sql : str
        SQL query to execute
    params : dict
        Placeholders (parameters) for templating. These are identified with ${placeholder} in the SQL script.

    Notes
    -----
    Any arbitrary kwargs passed to the class will be added to params.
    """

    def execute(self):
        self.output.df = self.spark.sql(self.query)



"""
Test the RegexpExtract and RegexpReplace columns transformations
"""

import pytest

from koheesio.logger import LoggingFactory
from koheesio.spark.transformations.strings.regexp import RegexpExtract, RegexpReplace

pytestmark = pytest.mark.spark

log = LoggingFactory.get_logger(name=__file__, inherit_from_koheesio=True)

data_year_wk = [["2020 W1"], ["2021 WK2"], ["2022WK3"]]
schema_year_wk = ["year_week"]
regexp_year_wk = "([0-9]{4}) ?WK?([0-9]+)"

data_with_strings = [[1, 2, "hello"], [3, 4, "world"], [3, 4, None]]
schema_with_strings = ["c1", "c2", "c3"]
regexp_with_strings = "([a-zA-Z]+)([0-9]+)"


@pytest.mark.parametrize(
    "input_values,expected",
    [
        (
            # description: Test RegexpExtract - index 0: should be the whole match
            # input values
            dict(column=["year_week"], target_column="result", regexp=regexp_year_wk),
            # expected data
            [
                dict(year_week="2020 W1", result="2020 W1"),
                dict(year_week="2021 WK2", result="2021 WK2"),
                dict(year_week="2022WK3", result="2022WK3"),
            ],
        ),
        (
            # description: Test RegexpExtract - index 1: should be the year in this example
            # input values
            dict(column=["year_week"], target_column="year", regexp=regexp_year_wk, index=1),
            # expected data
            [
                dict(year_week="2020 W1", year="2020"),
                dict(year_week="2021 WK2", year="2021"),
                dict(year_week="2022WK3", year="2022"),
            ],
        ),
        (
            # description: Test RegexpExtract - index 2: should be the week in this example
            # input values
            dict(column=["year_week"], target_column="week", regexp=regexp_year_wk, index=2),
            # expected data
            [
                dict(year_week="2020 W1", week="1"),
                dict(year_week="2021 WK2", week="2"),
                dict(year_week="2022WK3", week="3"),
            ],
        ),
    ],
)
def test_regexp_extract(input_values, expected, spark):
    input_df = spark.createDataFrame(data_year_wk, schema_year_wk)

    output_df = RegexpExtract(**input_values).transform(input_df)

    # log equivalent of doing df.show()
    log.info(f"show output_df: \n{output_df._jdf.showString(20, 20, False)}")
    actual = [row.asDict() for row in output_df.collect()]
    assert actual == expected


# TODO: not supported yet
#
#     def test_transform_nested_df(self) -> None:
#         """
#         Test with a nested dataframe
#         """
#         output_df = (
#             RegexpExtract(
#                 df=self.sample_nested_df,
#                 params={
#                     "source_column": "payload.col2.col3",
#                     "reg_exp": "([a-zA-Z]+)([0-9]+)",
#                     "index": 1,
#                 },
#                 target_alias="extracted",
#                 columns_to_flatten=["payload{}.col2{}"],
#             )
#             .execute()
#             .df
#         )
#
#         output_list = [row.extracted for row in output_df.collect()]
#         expected_output = ["value"]
#         self.assertEqual(output_list.sort(), expected_output.sort())


@pytest.mark.parametrize(
    "input_values,expected",
    [
        (
            # description: Test RegexpReplace - super simple regexp
            # input values
            dict(column=["c3"], target_column="replaced", regexp="l", replacement="-L-"),
            # expected data
            ["he-L--L-o", "wor-L-d", None],
        ),
        (
            # description: Test RegexpExtract - a bit more complex regexp
            # input values
            dict(column=["c3"], target_column="replaced", regexp="[ho]", replacement="*"),
            # expected data
            ["*ell*", "w*rld", None],
        ),
    ],
)
def test_regexp_replace(input_values, expected, spark):
    input_df = spark.createDataFrame(data_with_strings, schema_with_strings)

    regexp_replace = RegexpReplace(**input_values)
    target_column = regexp_replace.target_column
    output_df = regexp_replace.transform(input_df)

    # log equivalent of doing df.show()
    log.info(f"show output_df: \n{output_df._jdf.showString(20, 20, False)}")
    actual = [row.asDict()[target_column] for row in output_df.collect()]
    assert actual == expected


# TODO: not supported yet
#
#     def test_transform_nested_df(self) -> None:
#         """
#         Test with a nested dataframe
#         Replace a number with #NUMBER#
#         """
#         output_df = (
#             RegexpReplace(
#                 df=self.sample_nested_df,
#                 params={
#                     "source_column": "payload.col2.col3",
#                     "reg_exp": "[0-9]+",
#                     "replacement": "#NUMBER#",
#                 },
#                 target_alias="replaced",
#                 columns_to_flatten=["payload{}.col2{}"],
#             )
#             .execute()
#             .df
#         )
#
#         output_list = [row.replaced for row in output_df.collect()]
#         expected_output = ["value#NUMBER#"]
#         self.assertEqual(
#             output_list,
#             expected_output,
#         )



"""
This module defines the DeltaTableStreamWriter class, which is used to write streaming dataframes to Delta tables.
"""

from typing import Optional
from email.policy import default

from pydantic import Field

from koheesio.models import BaseModel
from koheesio.spark.writers.delta.batch import DeltaTableWriter
from koheesio.spark.writers.stream import StreamWriter


class DeltaTableStreamWriter(StreamWriter, DeltaTableWriter):
    """Delta table stream writer"""

    class Options(BaseModel):
        """Options for DeltaTableStreamWriter"""

        allow_population_by_field_name: bool = Field(
            default=True, description=" To do convert to Field and pass as .options(**config)"
        )
        maxBytesPerTrigger: Optional[str] = Field(
            default=None, description="How much data to be processed per trigger. The default is 1GB"
        )
        maxFilesPerTrigger: int = Field(
            default == 1000,
            description="The maximum number of new files to be considered in every trigger (default: 1000).",
        )

    def execute(self):
        if self.batch_function:
            self.streaming_query = self.writer.start()
        else:
            self.streaming_query = self.writer.toTable(tableName=self.table.table_name)



"""Read from a location using Databricks' `autoloader`

Autoloader can ingest JSON, CSV, PARQUET, AVRO, ORC, TEXT, and BINARYFILE file formats.
"""

from typing import Any, Dict, List, Optional, Tuple, Union
from enum import Enum

from pyspark.sql.streaming import DataStreamReader
from pyspark.sql.types import AtomicType, StructType

from koheesio.models import Field, field_validator
from koheesio.spark.readers import Reader


class AutoLoaderFormat(Enum):
    """The file format, used in `cloudFiles.format`
    Autoloader supports JSON, CSV, PARQUET, AVRO, ORC, TEXT, and BINARYFILE file formats.
    """

    JSON = "json"
    CSV = "csv"
    PARQUET = "parquet"
    AVRO = "avro"
    ORC = "orc"
    TEXT = "text"
    BINARYFILE = "binaryfile"


class AutoLoader(Reader):
    """Read from a location using Databricks' `autoloader`

    Autoloader can ingest JSON, CSV, PARQUET, AVRO, ORC, TEXT, and BINARYFILE file formats.

    Notes
    -----
    `autoloader` is a `Spark Structured Streaming` function!

    Although most transformations are compatible with `Spark Structured Streaming`, not all of them are. As a result,
    be mindful with your downstream transformations.

    Parameters
    ----------
    format : Union[str, AutoLoaderFormat]
        The file format, used in `cloudFiles.format`. Autoloader supports JSON, CSV, PARQUET, AVRO, ORC, TEXT, and
        BINARYFILE file formats.
    location : str
        The location where the files are located, used in `cloudFiles.location`
    schema_location : str
        The location for storing inferred schema and supporting schema evolution, used in `cloudFiles.schemaLocation`.
    options : Optional[Dict[str, str]], optional, default={}
        Extra inputs to provide to the autoloader. For a full list of inputs, see
        https://docs.databricks.com/ingestion/auto-loader/options.html


    Example
    -------
    ```python
    from koheesio.steps.readers.databricks import AutoLoader, AutoLoaderFormat

    result_df = AutoLoader(
        format=AutoLoaderFormat.JSON,
        location="some_s3_path",
        schema_location="other_s3_path",
        options={"multiLine": "true"},
    ).read()
    ```

    See Also
    --------
    Some other useful documentation:

    - autoloader: https://docs.databricks.com/ingestion/auto-loader/index.html
    - Spark Structured Streaming: https://spark.apache.org/docs/latest/structured-streaming-programming-guide.html
    """

    format: Union[str, AutoLoaderFormat] = Field(default=..., description=AutoLoaderFormat.__doc__)
    location: str = Field(
        default=...,
        description="The location where the files are located, used in `cloudFiles.location`",
    )
    schema_location: str = Field(
        default=...,
        alias="schemaLocation",
        description="The location for storing inferred schema and supporting schema evolution, "
        "used in `cloudFiles.schemaLocation`.",
    )
    options: Optional[Dict[str, Any]] = Field(
        default_factory=dict,
        description="Extra inputs to provide to the autoloader. For a full list of inputs, "
        "see https://docs.databricks.com/ingestion/auto-loader/options.html",
    )
    schema_: Optional[Union[str, StructType, List[str], Tuple[str, ...], AtomicType]] = Field(
        default=None,
        description="Explicit schema to apply to the input files.",
        alias="schema",
    )

    @field_validator("format")
    def validate_format(cls, format_specified):
        """Validate `format` value"""
        if isinstance(format_specified, str):
            if format_specified.upper() in [f.value.upper() for f in AutoLoaderFormat]:
                format_specified = getattr(AutoLoaderFormat, format_specified.upper())
        return str(format_specified.value)

    def get_options(self):
        """Get the options for the autoloader"""
        self.options.update(
            {
                "cloudFiles.format": self.format,
                "cloudFiles.schemaLocation": self.schema_location,
            }
        )
        return self.options

    # @property
    def reader(self) -> DataStreamReader:
        reader = self.spark.readStream.format("cloudFiles")
        if self.schema_ is not None:
            reader = reader.schema(self.schema_)
        reader = reader.options(**self.get_options())
        return reader

    def execute(self):
        """Reads from the given location with the given options using Autoloader"""
        self.output.df = self.reader().load(self.location)



import pytest

from koheesio.spark.writers.dummy import DummyWriter

pytestmark = pytest.mark.spark

expected = {"id": 0}


def test_write(dummy_df):
    test_writer = DummyWriter()
    actual = test_writer.write(dummy_df)
    assert actual.get("head") == expected


def test_write_unhappy():
    """Calling .write() when no DataFrame was passed or set should result in a RuntimeError"""
    test_writer = DummyWriter()
    with pytest.raises(RuntimeError):
        test_writer.write()


def test_execute(dummy_df):
    test_writer = DummyWriter(df=dummy_df)
    actual = test_writer.execute()
    assert actual.get("head") == expected



"""
Generic file Readers for different file formats.

Supported file formats:
- CSV
- Parquet
- Avro
- JSON
- ORC
- Text

Examples:
```python
from koheesio.spark.readers import (
    CsvReader,
    ParquetReader,
    AvroReader,
    JsonReader,
    OrcReader,
)

csv_reader = CsvReader(path="path/to/file.csv", header=True)
parquet_reader = ParquetReader(path="path/to/file.parquet")
avro_reader = AvroReader(path="path/to/file.avro")
json_reader = JsonReader(path="path/to/file.json")
orc_reader = OrcReader(path="path/to/file.orc")
```

For more information about the available options, see Spark's
[official documentation](https://spark.apache.org/docs/latest/sql-data-sources.html).
"""

from typing import Optional, Union
from enum import Enum
from pathlib import Path

from pyspark.sql.types import StructType

from koheesio.models import ExtraParamsMixin, Field, field_validator
from koheesio.spark.readers import Reader


class FileFormat(str, Enum):
    """Supported file formats.

    This enum represents the supported file formats that can be used with the FileLoader class.
    The available file formats are:
    - csv: Comma-separated values format
    - parquet: Apache Parquet format
    - avro: Apache Avro format
    - json: JavaScript Object Notation format
    - orc: Apache ORC format
    - text: Plain text format
    """

    csv = "csv"
    parquet = "parquet"
    avro = "avro"
    json = "json"
    orc = "orc"
    # excel = "excel"  # TODO: Add support for Excel
    # xml = "xml"  # TODO: Add support for XML
    # yaml = "yaml"  # TODO: Add support for YAML
    text = "text"


# pylint: disable=line-too-long
class FileLoader(Reader, ExtraParamsMixin):
    """Generic file reader.

    Available file formats:
    - CSV
    - Parquet
    - Avro
    - JSON
    - ORC
    - Text (default)

    Extra parameters can be passed to the reader using the `extra_params` attribute or as keyword arguments.

    Example:
    ```python
    reader = FileLoader(
        path="path/to/textfile.txt", format="text", header=True, lineSep="\n"
    )
    ```

    For more information about the available options, see Spark's
    [official pyspark documentation](https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrameReader.text.html)
    and [read about text data source](https://spark.apache.org/docs/latest/sql-data-sources-text.html).

    Also see the [data sources generic options](https://spark.apache.org/docs/3.5.0/sql-data-sources-generic-options.html).
    """

    format: FileFormat = Field(default=FileFormat.text, description="File format to read")
    path: Union[Path, str] = Field(default=..., description="Path to the file to read")
    schema_: Optional[Union[StructType, str]] = Field(
        default=None, description="Schema to use when reading the file", validate_default=False, alias="schema"
    )
    streaming: Optional[bool] = Field(default=False, description="Whether to read the files as a Stream or not")

    @field_validator("path")
    def ensure_path_is_str(cls, v):
        """Ensure that the path is a string as required by Spark."""
        if isinstance(v, Path):
            return str(v.absolute().as_posix())
        return v

    def execute(self):
        """Reads the file, in batch or as a stream, using the specified format and schema, while applying any extra parameters."""
        reader = self.spark.readStream if self.streaming else self.spark.read
        reader = reader.format(self.format)

        if self.schema_:
            reader.schema(self.schema_)

        if self.extra_params:
            reader = reader.options(**self.extra_params)

        self.output.df = reader.load(self.path)


class CsvReader(FileLoader):
    """Reads a CSV file.

    This class is a convenience class that sets the `format` field to `FileFormat.csv`.

    Extra parameters can be passed to the reader using the `extra_params` attribute or as keyword arguments.

    Example:
    ```python
    reader = CsvReader(path="path/to/file.csv", header=True)
    ```

    For more information about the available options,
    see the
    [official pyspark documentation](https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrameReader.csv.html)
    and [read about CSV data source](https://spark.apache.org/docs/latest/sql-data-sources-csv.html).

    Also see the
    [data sources generic options](https://spark.apache.org/docs/latest/sql-data-sources-generic-options.html).
    """

    format: FileFormat = FileFormat.csv


class ParquetReader(FileLoader):
    """Reads a Parquet file.

    This class is a convenience class that sets the `format` field to `FileFormat.parquet`.

    Extra parameters can be passed to the reader using the `extra_params` attribute or as keyword arguments.

    Example:
    ```python
    reader = ParquetReader(path="path/to/file.parquet", mergeSchema=True)
    ```

    For more information about the available options,
    see the
    [official pyspark documentation](https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrameReader.parquet.html)
    and [read about Parquet data source](https://spark.apache.org/docs/latest/sql-data-sources-parquet.html).

    Also see the
    [data sources generic options](https://spark.apache.org/docs/latest/sql-data-sources-generic-options.html).
    """

    format: FileFormat = FileFormat.parquet


class AvroReader(FileLoader):
    """Reads an Avro file.

    This class is a convenience class that sets the `format` field to `FileFormat.avro`.

    Extra parameters can be passed to the reader using the `extra_params` attribute or as keyword arguments.

    Example:
    ```python
    reader = AvroReader(path="path/to/file.avro", mergeSchema=True)
    ```

    Make sure to have the `spark-avro` package installed in your environment.

    For more information about the available options,
    see the [official documentation](https://spark.apache.org/docs/latest/sql-data-sources-avro.html#content).
    """

    format: FileFormat = FileFormat.avro


class JsonReader(FileLoader):
    """Reads a JSON file.

    This class is a convenience class that sets the `format` field to `FileFormat.json`.

    Extra parameters can be passed to the reader using the `extra_params` attribute or as keyword arguments.

    Example:
    ```python
    reader = JsonReader(path="path/to/file.json", allowComments=True)
    ```

    For more information about the available options,
    see the
    [official pyspark documentation](https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrameReader.json.html#pyspark.sql.DataFrameReader.json)
    and [read about JSON data source](https://spark.apache.org/docs/latest/sql-data-sources-json.html).

    Also see the
    [data sources generic options](https://spark.apache.org/docs/latest/sql-data-sources-generic-options.html).
    """

    format: FileFormat = FileFormat.json


class OrcReader(FileLoader):
    """Reads an ORC file.

    This class is a convenience class that sets the `format` field to `FileFormat.orc`.

    Extra parameters can be passed to the reader using the `extra_params` attribute or as keyword arguments.

    Example:
    ```python
    reader = OrcReader(path="path/to/file.orc", mergeSchema=True)
    ```

    For more information about the available options,
    see the
    [official documentation](https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrameReader.orc.html)
    and [read about ORC data source](https://spark.apache.org/docs/latest/sql-data-sources-orc.html).

    Also see the
    [data sources generic options](https://spark.apache.org/docs/latest/sql-data-sources-generic-options.html).
    """

    format: FileFormat = FileFormat.orc



"""
This module contains a few simple HTTP Steps that can be used to perform API Calls to HTTP endpoints

Example
-------
```python
from koheesio.steps.http import HttpGetStep

response = HttpGetStep(url="https://google.com").execute().json_payload
```

In the above example, the `response` variable will contain the JSON response from the HTTP request.
"""

import json
from typing import Any, Dict, List, Optional, Union
from enum import Enum

import requests

from koheesio import Step
from koheesio.models import (
    ExtraParamsMixin,
    Field,
    SecretStr,
    field_serializer,
    field_validator,
)

__all__ = [
    "HttpMethod",
    "HttpStep",
    "HttpGetStep",
    "HttpPostStep",
    "HttpPutStep",
    "HttpDeleteStep",
    "PaginatedHtppGetStep",
]


class HttpMethod(str, Enum):
    """
    Enumeration of allowed http methods
    """

    GET = "get"
    POST = "post"
    PUT = "put"
    DELETE = "delete"

    @classmethod
    def from_string(cls, value: str):
        """Allows for getting the right Method Enum by simply passing a string value
        This method is not case-sensitive
        """
        return getattr(cls, value.upper())


class HttpStep(Step, ExtraParamsMixin):
    """
    Can be used to perform API Calls to HTTP endpoints

    Understanding Retries
    ----------------------

    This class includes a built-in retry mechanism for handling temporary issues, such as network errors or server
    downtime, that might cause the HTTP request to fail. The retry mechanism is controlled by three parameters:
    `max_retries`, `initial_delay`, and `backoff`.

    - `max_retries` determines the number of retries after the initial request. For example, if `max_retries` is set to
        4, the request will be attempted a total of 5 times (1 initial attempt + 4 retries). If `max_retries` is set to
        0, no retries will be attempted, and the request will be tried only once.

    - `initial_delay` sets the waiting period before the first retry. If `initial_delay` is set to 3, the delay before
        the first retry will be 3 seconds. Changing the `initial_delay` value directly affects the amount of delay
        before each retry.

    - `backoff` controls the rate at which the delay increases for each subsequent retry. If `backoff` is set to 2 (the
        default), the delay will double with each retry. If `backoff` is set to 1, the delay between retries will
        remain constant. Changing the `backoff` value affects how quickly the delay increases.

    Given the default values of `max_retries=3`, `initial_delay=2`, and `backoff=2`, the delays between retries would
    be 2 seconds, 4 seconds, and 8 seconds, respectively. This results in a total delay of 14 seconds before all
    retries are exhausted.

    For example, if you set `initial_delay=3` and `backoff=2`, the delays before the retries would be `3 seconds`,
    `6 seconds`, and `12 seconds`. If you set `initial_delay=2` and `backoff=3`, the delays before the retries would be
    `2 seconds`, `6 seconds`, and `18 seconds`. If you set `initial_delay=2` and `backoff=1`, the delays before the
    retries would be `2 seconds`, `2 seconds`, and `2 seconds`.
    """

    url: str = Field(
        default=...,
        description="API endpoint URL",
        alias="uri",
    )
    headers: Optional[Dict[str, Union[str, SecretStr]]] = Field(
        default_factory=dict,
        description="Request headers",
        alias="header",
    )
    data: Optional[Union[Dict[str, str], str]] = Field(
        default_factory=dict, description="[Optional] Data to be sent along with the request", alias="body"
    )
    params: Optional[Dict[str, Any]] = Field(
        default_factory=dict,
        description="[Optional] Set of extra parameters that should be passed to HTTP request",
    )
    timeout: Optional[int] = Field(default=3, description="[Optional] Request timeout")
    method: Union[str, HttpMethod] = Field(
        default=HttpMethod.GET,
        description="What type of Http call to perform. One of 'get', 'post', 'put', 'delete'. Defaults to 'get'.",
    )
    session: requests.Session = Field(
        default_factory=requests.Session,
        description="Requests session object to be used for making HTTP requests",
        exclude=True,
        repr=False,
    )

    class Output(Step.Output):
        """Output class for HttpStep"""

        response_raw: Optional[requests.Response] = Field(
            default=None,
            alias="response",
            description="The raw requests.Response object returned by the appropriate requests.request() call",
        )
        response_json: Optional[Union[Dict, List]] = Field(
            default=None, alias="json_payload", description="The JSON response for the request"
        )
        raw_payload: Optional[str] = Field(
            default=None, alias="response_text", description="The raw response for the request"
        )
        status_code: Optional[int] = Field(default=None, description="The status return code of the request")

        @property
        def json_payload(self):
            """Alias for response_json"""
            return self.response_json

    @field_validator("method")
    def get_proper_http_method_from_str_value(cls, method_value):
        """Converts string value to HttpMethod enum value"""
        if isinstance(method_value, str):
            try:
                method_value = HttpMethod.from_string(method_value)
            except AttributeError as e:
                raise AttributeError(
                    "Only values from HttpMethod class are allowed! "
                    f"Provided value: '{method_value}', allowed values: {', '.join(HttpMethod.__members__.keys())}"
                ) from e

        return method_value

    @field_validator("headers", mode="before")
    def encode_sensitive_headers(cls, headers):
        """
        Encode potentially sensitive data into pydantic.SecretStr class to prevent them
        being displayed as plain text in logs.
        """
        if auth := headers.get("Authorization"):
            headers["Authorization"] = auth if isinstance(auth, SecretStr) else SecretStr(auth)
        return headers

    @field_serializer("headers", when_used="json")
    def decode_sensitive_headers(self, headers):
        """
        Authorization headers are being converted into SecretStr under the hood to avoid dumping any
        sensitive content into logs by the `encode_sensitive_headers` method.

        However, when calling the `get_headers` method, the SecretStr should be converted back to
        string, otherwise sensitive info would have looked like '**********'.

        This method decodes values of the `headers` dictionary that are of type SecretStr into plain text.
        """
        for k, v in headers.items():
            headers[k] = v.get_secret_value() if isinstance(v, SecretStr) else v
        return headers

    def get_headers(self):
        """
        Dump headers into JSON without SecretStr masking.
        """
        return json.loads(self.model_dump_json()).get("headers")

    def set_outputs(self, response):
        """
        Types of response output
        """
        self.output.response_raw = response
        self.output.raw_payload = response.text
        self.output.status_code = response.status_code

        # Only decode non empty payloads to avoid triggering decoding error unnecessarily.
        if self.output.raw_payload:
            try:
                self.output.response_json = response.json()

            except json.decoder.JSONDecodeError as e:
                self.log.info(f"An error occurred while processing the JSON payload. Error message:\n{e.msg}")

    def get_options(self):
        """options to be passed to requests.request()"""
        return {
            "url": self.url,
            "headers": self.get_headers(),
            "data": self.data,
            "timeout": self.timeout,
            **self.params,  # type: ignore
        }

    def request(self, method: Optional[HttpMethod] = None) -> requests.Response:
        """
        Executes the HTTP request with retry logic.

        Actual http_method execution is abstracted into this method.
        This is to avoid unnecessary code duplication. Allows to centrally log, set outputs, and validated.

        This method will try to execute `requests.request` up to `self.max_retries` times. If `self.request()` raises
        an exception, it logs a warning message and the error message, then waits for
        `self.initial_delay * (self.backoff ** i)` seconds before retrying. The delay increases exponentially
        after each failed attempt due to the `self.backoff ** i` term.

        If `self.request()` still fails after `self.max_retries` attempts, it logs an error message and re-raises the
        last exception that was caught.

        This is a good way to handle temporary issues that might cause `self.request()` to fail, such as network errors
        or server downtime. The exponential backoff ensures that you're not constantly bombarding a server with
        requests if it's struggling to respond.

        Parameters
        ----------
        method : HttpMethod
            Optional parameter that allows calls to different HTTP methods and bypassing class level `method`
            parameter.

        Raises
        ------
        requests.RequestException, requests.HTTPError
            The last exception that was caught if `requests.request()` fails after `self.max_retries` attempts.
        """
        _method = (method or self.method).value.upper()
        options = self.get_options()

        self.log.debug(f"Making {_method} request to {options['url']} with headers {options['headers']}")

        response = self.session.request(method=_method, **options)
        response.raise_for_status()

        self.log.debug(f"Received response with status code {response.status_code} and body {response.text}")
        self.set_outputs(response)

        return response

    def get(self) -> requests.Response:
        """Execute an HTTP GET call"""
        self.method = HttpMethod.GET
        return self.request()

    def post(self) -> requests.Response:
        """Execute an HTTP POST call"""
        self.method = HttpMethod.POST
        return self.request()

    def put(self) -> requests.Response:
        """Execute an HTTP PUT call"""
        self.method = HttpMethod.PUT
        return self.request()

    def delete(self) -> requests.Response:
        """Execute an HTTP DELETE call"""
        self.method = HttpMethod.DELETE
        return self.request()

    def execute(self) -> Output:
        """
        Executes the HTTP request.

        This method simply calls `self.request()`, which includes the retry logic. If `self.request()` raises an
        exception, it will be propagated to the caller of this method.

        Raises
        ------
        requests.RequestException, requests.HTTPError
            The last exception that was caught if `self.request()` fails after `self.max_retries` attempts.
        """
        self.request()


class HttpGetStep(HttpStep):
    """send GET requests

    Example
    -------
    ```python
    response = HttpGetStep(url="https://google.com").execute().json_payload
    ```
    In the above example, the `response` variable will contain the JSON response from the HTTP request.
    """

    method: HttpMethod = HttpMethod.GET


class HttpPostStep(HttpStep):
    """send POST requests"""

    method: HttpMethod = HttpMethod.POST


class HttpPutStep(HttpStep):
    """send PUT requests"""

    method: HttpMethod = HttpMethod.PUT


class HttpDeleteStep(HttpStep):
    """send DELETE requests"""

    method: HttpMethod = HttpMethod.DELETE


class PaginatedHtppGetStep(HttpGetStep):
    """
    Represents a paginated HTTP GET step.

    Parameters
    ----------
    paginate : bool, optional
        Whether to paginate the API response. Defaults to False.
    pages : int, optional
        Number of pages to paginate. Defaults to 1.
    offset : int, optional
        Offset for paginated API calls. Offset determines the starting page. Defaults to 1.
    limit : int, optional
        Limit for paginated API calls. Defaults to 100.
    """

    paginate: Optional[bool] = Field(
        default=False,
        description="Whether to paginate the API response. Defaults to False. When set to True, the API response will "
        "be paginated. The url should contain a named 'page' parameter for example: "
        "api.example.com/data?page={page}",
    )
    pages: Optional[int] = Field(default=1, description="Number of pages to paginate. Defaults to 1")
    offset: Optional[int] = Field(
        default=1,
        description="Offset for paginated API calls. Offset determines the starting page. Defaults to 1. "
        "The url can (optionally) contain a named 'offset' parameter, for example: "
        "api.example.com/data?offset={offset}",
    )
    limit: Optional[int] = Field(
        default=100,
        description="Limit for paginated API calls. The url should (optionally) contain a named limit parameter, "
        "for example: api.example.com/data?limit={limit}",
    )

    def _adjust_params(self) -> Dict[str, Any]:
        """
        Adjusts the parameters by removing the 'paginate' key.

        Returns
        -------
        dict
            The adjusted parameters.
        """
        return {k: v for k, v in self.params.items() if k not in ["paginate"]}  # type: ignore

    def get_options(self):
        """
        Returns the options to be passed to the requests.request() function.

        Returns
        -------
        dict
            The options.
        """
        options = {
            "url": self.url,
            "headers": self.get_headers(),
            "data": self.data,
            "timeout": self.timeout,
            **self._adjust_params(),  # type: ignore
        }

        return options

    def _url(self, basic_url: str, page: Optional[int] = None) -> str:
        """
        Adds additional parameters to the URL.

        Parameters
        ----------
        basic_url : str
            The basic URL without any parameters.
        page : int, optional
            The page number for pagination. Defaults to None.

        Returns
        -------
        str
            The URL with additional parameters.
        """
        url_params = self._adjust_params()

        if self.paginate:
            url_params.update(
                {
                    "offset": self.offset,
                    "limit": self.limit,
                    "page": page,
                }
            )

        return basic_url.format(**url_params)

    def execute(self) -> HttpGetStep.Output:
        """
        Executes the HTTP GET request and handles pagination.

        Returns
        -------
        HttpGetStep.Output
            The output of the HTTP GET request.
        """
        # Set up pagination parameters
        offset, pages = (self.offset, self.pages + 1) if self.paginate else (1, 1)  # type: ignore
        data = []
        _basic_url = self.url

        for page in range(offset, pages):
            if self.paginate:
                self.log.info(f"Fetching page {page} of {pages - 1}")

            self.url = self._url(basic_url=_basic_url, page=page)
            self.request()

            if isinstance(self.output.response_json, list):
                data += self.output.response_json
            else:
                data.append(self.output.response_json)

        self.url = _basic_url
        self.output.response_json = data
        self.output.response_raw = None
        self.output.raw_payload = None
        self.output.status_code = None



"""Dummy transformation for testing purposes."""

from pyspark.sql.functions import lit

from koheesio.spark.transformations import Transformation


class DummyTransformation(Transformation):
    """Dummy transformation for testing purposes.

    This transformation adds a new column `hello` to the DataFrame with the value `world`.

    It is intended for testing purposes or for use in examples or reference documentation.

    Example
    -------
    __input_df:__

    | id |
    |----|
    | 1  |

    ```python
    output_df = DummyTransformation().transform(input_df)
    ```

    __output_df:__

    | id | hello |
    |----|-------|
    | 1  | world |

    In this example, the `hello` column is added to the DataFrame `input_df`.

    """

    def execute(self):
        self.output.df = self.df.withColumn("hello", lit("world"))



from typing import List, Union

import pytest

import pyspark
from pyspark.sql import SparkSession

from koheesio.utils import get_project_root

PROJECT_ROOT = get_project_root()

pytestmark = pytest.mark.spark

if pyspark.__version__.startswith("3.5"):
    pytestmark = pytest.mark.skip("Spark Expectations is not supported for Spark 3.5")


class TestSparkExpectationsTransform:
    """
    Test the transform step in isolation by providing sample df and rules. Will check if rows are dropped and written
    to err_table, and also if meta columns are dropped.

    dq rules for test case:
    - col1 should not be null
    - col2 should not be null
    """

    input_data = [
        ["col1-1", "col2-1", "col3-1"],
        ["col1-2", "col2-2", "col3-2"],
        ["col1-3", None, "col3-3"],
        [None, "col2-4", "col3-4"],
    ]
    input_schema = ["col1", "col2", "col3"]
    expected = [
        {"col1": "col1-1", "col2": "col2-1", "col3": "col3-1"},
        {"col1": "col1-2", "col2": "col2-2", "col3": "col3-2"},
    ]

    @pytest.fixture(scope="session")
    def prepare_tables(self, spark, data_path):
        TestSparkExpectationsTransform.apply_spark_sql(
            spark,
            [
                f"{data_path}/transformations/spark_expectations_resources/test_product_dq_stats.ddl",
                f"{data_path}/transformations/spark_expectations_resources/test_product_rules.ddl",
                f"{data_path}/transformations/spark_expectations_resources/test_product_rules.sql",
            ],
        )

    def test_rows_are_dropped(self, spark: SparkSession, prepare_tables):
        from koheesio.integrations.spark.dq.spark_expectations import (
            SparkExpectationsTransformation,
        )

        input_df = spark.createDataFrame(
            TestSparkExpectationsTransform.input_data,
            schema=TestSparkExpectationsTransform.input_schema,
        )

        se_transform = SparkExpectationsTransformation(
            product_id="test_product",
            product_rules_table="default.test_product_rules",
            dq_stats_table_name="default.test_product_dq_stats",
            target_table_name="default.output_table",
        )

        output_df = se_transform.transform(input_df)

        actual = [k.asDict() for k in output_df.select("col1", "col2", "col3").collect()]
        assert actual == TestSparkExpectationsTransform.expected

        err_table_df = spark.read.table("default.output_table_error")
        assert err_table_df.count() == 2

        spark.sql("drop table default.output_table_error")

    def test_meta_columns_are_not_dropped(self, spark, prepare_tables):
        from koheesio.integrations.spark.dq.spark_expectations import (
            SparkExpectationsTransformation,
        )

        input_df = spark.createDataFrame(
            TestSparkExpectationsTransform.input_data,
            schema=TestSparkExpectationsTransform.input_schema,
        )

        se_transform = SparkExpectationsTransformation(
            product_id="test_product",
            product_rules_table="default.test_product_rules",
            dq_stats_table_name="default.test_product_dq_stats",
            target_table_name="default.output_table",
            drop_meta_column=False,
        )

        output_df = se_transform.transform(input_df)
        output_columns = output_df.columns

        assert "meta_dq_run_id" in output_columns
        # Spark Expectations should add either meta_dq_run_date (<=1.1.0) or meta_dq_run_datetime (>= v1.1.1)
        assert "meta_dq_run_date" in output_columns or "meta_dq_run_datetime" in output_columns

        spark.sql("drop table default.output_table_error")

    def test_meta_columns_are_dropped(self, spark, prepare_tables):
        from koheesio.integrations.spark.dq.spark_expectations import (
            SparkExpectationsTransformation,
        )

        input_df = spark.createDataFrame(
            TestSparkExpectationsTransform.input_data,
            schema=TestSparkExpectationsTransform.input_schema,
        )

        se_transform = SparkExpectationsTransformation(
            product_id="test_product",
            product_rules_table="default.test_product_rules",
            dq_stats_table_name="default.test_product_dq_stats",
            target_table_name="default.output_table",
            drop_meta_column=True,
        )

        output_df = se_transform.transform(input_df)
        output_columns = output_df.columns

        assert "meta_dq_run_id" not in output_columns
        assert "meta_dq_run_datetime" not in output_columns and "meta_dq_run_datetime" not in output_columns

        spark.sql("drop table default.output_table_error")

    @staticmethod
    def apply_spark_sql(spark: SparkSession, source_sql_files: Union[List[str], str]) -> None:
        if isinstance(source_sql_files, str):
            source_sql_files = [source_sql_files]

        for source_sql_file in source_sql_files:
            with open(source_sql_file, "r") as f:
                content = f.read()
                spark.sql(content)

    def test_with_full_se_user_conf(self):
        from koheesio.integrations.spark.dq.spark_expectations import (
            SparkExpectationsTransformation,
        )

        conf = {
            "spark.expectations.notifications.email.enabled": False,
            "spark.expectations.notifications.email.smtp_host": "mailhost.email.com",
            "spark.expectations.notifications.email.smtp_port": 25,
            "spark.expectations.notifications.email.from": "",
            "spark.expectations.notifications.email.to.other.mail.com": "",
            "spark.expectations.notifications.email.subject": "{} - data quality - notifications",
            "spark.expectations.notifications.slack.enabled": True,
            "spark.expectations.notifications.slack.webhook_url": "https://hooks.slack.com/...",
            "spark.expectations.notifications.on_start": True,
            "spark.expectations.notifications.on_completion": True,
            "spark.expectations.notifications.on_fail": True,
            "spark.expectations.notifications.on.error.drop.exceeds.threshold.breach": False,
            "spark.expectations.notifications.error.drop.threshold": 15,
        }
        instance = SparkExpectationsTransformation(
            product_id="test_product",
            product_rules_table="default.test_product_rules",
            dq_stats_table_name="default.test_product_dq_stats",
            target_table_name="default.output_table",
            drop_meta_column=True,
            se_user_conf=conf,
        )
        assert instance.se_user_conf == conf

    def test_overwrite_error_writer(self):
        from koheesio.integrations.spark.dq.spark_expectations import (
            SparkExpectationsTransformation,
        )

        """
        Test that the error_writer can be overwritten using error_writer_mode and error_writer_format
        """
        se_transform = SparkExpectationsTransformation(
            product_id="test_product",
            rules_table="default.test_product_rules",
            statistics_table="default.test_product_dq_stats",
            target_table="default.output_table",
            drop_meta_column=True,
            se_user_conf={},
            mode="append",
            format="delta",
            error_writer_mode="overwrite",
            error_writer_format="parquet",
            error_writing_options={"mergeSchema": "true"},
        )
        error_writer = se_transform._error_table_writer
        stats_writer = se_transform._stats_table_writer
        assert error_writer._mode == "overwrite"
        assert error_writer._format == "parquet"
        assert stats_writer._mode == "append"
        assert stats_writer._format == "delta"
        assert error_writer._options == {"mergeSchema": "true"}

    def test_overwrite_stats_writer(self):
        from koheesio.integrations.spark.dq.spark_expectations import (
            SparkExpectationsTransformation,
        )

        """
        Test that the stats_writer can be overwritten using stats_writer_mode and stats_writer_format and that the
        error_writer_options default = {}.
        """
        se_transform = SparkExpectationsTransformation(
            product_id="test_product",
            rules_table="default.test_product_rules",
            statistics_table="default.test_product_dq_stats",
            target_table="default.output_table",
            drop_meta_column=True,
            se_user_conf={},
            mode="append",
            format="delta",
            stats_writer_mode="overwrite",
            stats_writer_format="parquet",
        )
        error_writer = se_transform._error_table_writer
        stats_writer = se_transform._stats_table_writer
        assert error_writer._mode == "append"
        assert error_writer._format == "delta"
        assert error_writer._options == {}
        assert stats_writer._mode == "overwrite"
        assert stats_writer._format == "parquet"



"""Transformation to replace a particular value in a column with another one"""

from typing import Optional, Union

from pyspark.sql import Column
from pyspark.sql.functions import col, lit, when

from koheesio.models import Field
from koheesio.spark.transformations import ColumnsTransformationWithTarget
from koheesio.spark.utils import SparkDatatype


def replace(column: Union[Column, str], to_value: str, from_value: Optional[str] = None):
    """Function to replace a particular value in a column with another one"""
    # make sure we have a Column object
    if isinstance(column, str):
        column = col(column)

    if not from_value:
        condition = column.isNull()
    else:
        condition = column == from_value

    return when(condition, lit(to_value)).otherwise(column)


class Replace(ColumnsTransformationWithTarget):
    """
    Replace a particular value in a column with another one

    Can handle empty strings ("") as well as NULL / None values.

    Unsupported datatypes:
    ----------------------
    Following casts are not supported

    will raise an error in Spark:

    * binary
    * boolean
    * array<...>
    * map<...,...>

    Supported datatypes:
    --------------------
    Following casts are supported:

    * byte
    * short
    * integer
    * long
    * float
    * double
    * decimal
    * timestamp
    * date
    * string
    * void
        skipped by default

    Any supported none-string datatype will be cast to string before the replacement is done.

    Example
    -------
    __input_df__:

    | id | string |
    |----|--------|
    | 1  | hello  |
    | 2  | world  |
    | 3  |        |

    ```python
    output_df = Replace(
        column="string",
        from_value="hello",
        to_value="programmer",
    ).transform(input_df)
    ```

    __output_df__:

    | id |      string |
    |----|-------------|
    | 1  | programmer  |
    | 2  | world       |
    | 3  |             |

    In this example, the value "hello" in the column "string" is replaced with "programmer".
    """

    class ColumnConfig(ColumnsTransformationWithTarget.ColumnConfig):
        """Column type configurations for the column to be replaced"""

        run_for_all_data_type = [
            SparkDatatype.BYTE,
            SparkDatatype.SHORT,
            SparkDatatype.INTEGER,
            SparkDatatype.LONG,
            SparkDatatype.FLOAT,
            SparkDatatype.DOUBLE,
            SparkDatatype.DECIMAL,
            SparkDatatype.STRING,
            SparkDatatype.TIMESTAMP,
            SparkDatatype.DATE,
        ]
        limit_data_type = [
            *run_for_all_data_type,
            SparkDatatype.VOID,
        ]

    from_value: Optional[str] = Field(
        default=None,
        alias="from",
        description="The original value that needs to be replaced. If no value is given, all 'null' values will be "
        "replaced with the to_value",
    )
    to_value: str = Field(default=..., alias="to", description="The new value to replace this with")

    def func(self, column: Column) -> Column:
        return replace(column=column, from_value=self.from_value, to_value=self.to_value)



"""Ability to generate UUID5 using native pyspark (no udf)"""

import uuid
from typing import Optional, Union

from pyspark.sql import functions as f

from koheesio.models import Field, ListOfColumns, field_validator
from koheesio.spark.transformations import Transformation


def uuid5_namespace(ns: Optional[Union[str, uuid.UUID]]) -> uuid.UUID:
    """Helper function used to provide a UUID5 hashed namespace based on the passed str

    Parameters
    ----------
    ns : Optional[Union[str, uuid.UUID]]
        A str, an empty string (or None), or an existing UUID can be passed

    Returns
    -------
    uuid.UUID
        UUID5 hashed namespace
    """
    # if we already have a UUID, we just return it
    if isinstance(ns, uuid.UUID):
        return ns

    # if ns is empty or none, we simply return the default NAMESPACE_DNS
    if not ns:
        ns = uuid.NAMESPACE_DNS
        return ns

    # else we hash the string against the NAMESPACE_DNS
    ns = uuid.uuid5(uuid.NAMESPACE_DNS, ns)
    return ns


def hash_uuid5(
    input_value: str,
    namespace: Optional[Union[str, uuid.UUID]] = "",
    extra_string: Optional[str] = "",
):
    """pure python implementation of HashUUID5

    See: https://docs.python.org/3/library/uuid.html#uuid.uuid5

    Parameters
    ----------
    input_value : str
        value that will be hashed
    namespace : Optional[str | uuid.UUID]
        namespace DNS
    extra_string : Optional[str]
        optional extra string that will be prepended to the input_value

    Returns
    -------
    str
        uuid.UUID (uuid5) cast to string
    """
    if not isinstance(namespace, uuid.UUID):
        hashed_namespace = uuid5_namespace(namespace)
    else:
        hashed_namespace = namespace
    return str(uuid.uuid5(hashed_namespace, (extra_string + input_value)))


class HashUUID5(Transformation):
    """Generate a UUID with the UUID5 algorithm

    Spark does not provide inbuilt API to generate version 5 UUID, hence we have to use a custom implementation
    to provide this capability.

    Prerequisites: this function has no side effects. But be aware that in most cases, the expectation is that your
    data is clean (e.g. trimmed of leading and trailing spaces)

    Concept
    -------
    UUID5 is based on the SHA-1 hash of a namespace identifier (which is a UUID) and a name (which is a string).
    https://docs.python.org/3/library/uuid.html#uuid.uuid5

    Based on https://github.com/MrPowers/quinn/pull/96 with the difference that since Spark 3.0.0 an OVERLAY function
    from ANSI SQL 2016 is available which saves coding space and string allocation(s) in place of CONCAT + SUBSTRING.

    For more info on OVERLAY, see: https://docs.databricks.com/sql/language-manual/functions/overlay.html

    Example
    -------
    Input is a DataFrame with two columns:

    | id | string |
    |----|--------|
    | 1  | hello  |
    | 2  | world  |
    | 3  |        |

    Input parameters:

    - source_columns = ["id", "string"]
    - target_column = "uuid5"

    Result:

    | id | string | uuid5                                |
    |----|--------| ------------------------------------ |
    | 1  | hello  | f3e99bbd-85ae-5dc3-bf6e-cd0022a0ebe6 |
    | 2  | world  | b48e880f-c289-5c94-b51f-b9d21f9616c0 |
    | 3  |        | 2193a99d-222e-5a0c-a7d6-48fbe78d2708 |

    In code:
    ```python
    HashUUID5(source_columns=["id", "string"], target_column="uuid5").transform(
        input_df
    )
    ```

    In this example, the `id` and `string` columns are concatenated and hashed using the UUID5 algorithm. The result is
    stored in the `uuid5` column.
    """

    target_column: str = Field(
        default=..., description="The generated UUID will be written to the column name specified here"
    )
    source_columns: ListOfColumns = Field(
        default=...,
        description="List of columns that should be hashed. Should contain the name of at least 1 column. A list of "
        "columns or a single column can be specified. For example: `['column1', 'column2']` or `'column1'`",
    )
    delimiter: Optional[str] = Field(default="|", description="Separator for the string that will eventually be hashed")
    namespace: Optional[Union[str, uuid.UUID]] = Field(default="", description="Namespace DNS")
    extra_string: Optional[str] = Field(
        default="",
        description="In case of collisions, one can pass an extra string to hash on.",
    )

    # setting a shorter description to avoid excessive logging output
    description: str = "Generate a UUID with the UUID5 algorithm"

    @field_validator("source_columns")
    def _set_columns(cls, columns):
        """Ensures every column is wrapped in backticks"""
        columns = [f"`{column}`" for column in columns]
        return columns

    def execute(self) -> None:
        ns = f.lit(uuid5_namespace(self.namespace).bytes)
        self.log.info(f"UUID5 namespace '{ns}' derived from '{self.namespace}'")
        cols_to_hash = f.concat_ws(self.delimiter, *self.source_columns)
        cols_to_hash = f.concat(f.lit(self.extra_string), cols_to_hash)
        cols_to_hash = f.encode(cols_to_hash, "utf-8")
        cols_to_hash = f.concat(ns, cols_to_hash)
        source_columns_sha1 = f.sha1(cols_to_hash)
        variant_part = f.substring(source_columns_sha1, 17, 4)
        variant_part = f.conv(variant_part, 16, 2)
        variant_part = f.lpad(variant_part, 16, "0")
        variant_part = f.overlay(variant_part, f.lit("10"), 1, 2)  # RFC 4122 variant.
        variant_part = f.lower(f.conv(variant_part, 2, 16))
        target_col_uuid = f.concat_ws(
            "-",
            f.substring(source_columns_sha1, 1, 8),
            f.substring(source_columns_sha1, 9, 4),
            f.concat(f.lit("5"), f.substring(source_columns_sha1, 14, 3)),  # Set version.
            variant_part,
            f.substring(source_columns_sha1, 21, 12),
        )
        # Applying the transformation to the input df, storing the result in the column specified in `target_column`.
        self.output.df = self.df.withColumn(self.target_column, target_col_uuid)



import os
from unittest.mock import patch

import pytest

from pyspark.sql.types import StringType, StructField, StructType

from koheesio.spark.utils import on_databricks, schema_struct_to_schema_str
from koheesio.utils import get_args_for_func, get_random_string


def test_get_args_for_func():
    def func(a, b, c):
        return a + b + c

    func, args = get_args_for_func(func, {"a": 1, "b": 2, "c": 3})
    assert args == {"a": 1, "b": 2, "c": 3}


def test_import_class():
    import datetime

    from koheesio.utils import import_class

    assert import_class("datetime.datetime") == datetime.datetime


def test_get_random_string():
    assert get_random_string(10) != get_random_string(10)
    assert len(get_random_string(10)) == 10
    assert len(get_random_string(10, "abc")) == 10
    assert get_random_string(10, "abc").startswith("abc_")



"""
Convert the case of a string column to upper case, lower case, or title case

Classes
-------
`Lower`
    Converts a string column to lower case.
`Upper`
    Converts a string column to upper case.
`TitleCase` or `InitCap`
    Converts a string column to title case, where each word starts with a capital letter.
"""

from pyspark.sql import Column
from pyspark.sql.functions import initcap, lower, upper

from koheesio.spark.transformations import ColumnsTransformationWithTarget
from koheesio.spark.utils import SparkDatatype

__all__ = ["LowerCase", "UpperCase", "TitleCase", "InitCap"]


class LowerCase(ColumnsTransformationWithTarget):
    """
    This function makes the contents of a column lower case.

    Wraps the `pyspark.sql.functions.lower` function.

    Warnings
    --------
    If the type of the column is not string, `LowerCase` will not be run. A Warning will be thrown indicating this.

    Parameters
    ----------
    columns : Union[str, List[str]]
        The name of the column or columns to convert to lower case. Alias: column.
        Lower case will be applied to all columns in the list.
        Column is required to be of string type.
    target_column : str, optional
        The name of the column to store the result in. If None, the result will be stored in the same column as the
        input.

    Example
    -------
    __input_df:__

    |product            |amount|country|
    |-------------------|------|-------|
    |Banana lemon orange|  1000|    USA|
    |Carrots Blueberries|  1500|    USA|
    |              Beans|  1600|    USA|

    ```python
    output_df = LowerCase(
        column="product", target_column="product_lower"
    ).transform(df)
    ```

    __output_df:__

    |product            |amount|country|product_lower      |
    |-------------------|------|-------|-------------------|
    |Banana lemon orange|  1000|    USA|banana lemon orange|
    |Carrots Blueberries|  1500|    USA|carrots blueberries|
    |              Beans|  1600|    USA|              beans|

    In this example, the column `product` is converted to `product_lower` and the contents of this column are converted
    to lower case.
    """

    class ColumnConfig(ColumnsTransformationWithTarget.ColumnConfig):
        """Limit data type to string"""

        run_for_all_data_type = [SparkDatatype.STRING]
        limit_data_type = [SparkDatatype.STRING]

    def func(self, column: Column):
        return lower(column)


class UpperCase(LowerCase):
    """
    This function makes the contents of a column upper case.

    Wraps the `pyspark.sql.functions.upper` function.

    Warnings
    --------
    If the type of the column is not string, `UpperCase` will not be run. A Warning will be thrown indicating this.

    Parameters
    ----------
    columns : Union[str, List[str]]
        The name of the column or columns to convert to upper case. Alias: column.
        Upper case will be applied to all columns in the list.
        Column is required to be of string type.
    target_column : str, optional
        The name of the column to store the result in. If None, the result will be stored in the same column as the
        input.

    Examples
    -------
    __input_df:__

    |product            |amount|country|
    |-------------------|------|-------|
    |Banana lemon orange|  1000|    USA|
    |Carrots Blueberries|  1500|    USA|
    |              Beans|  1600|    USA|

    ```python
    output_df = UpperCase(
        column="product", target_column="product_upper"
    ).transform(df)
    ```

    __output_df:__

    |product            |amount|country|product_upper      |
    |-------------------|------|-------|-------------------|
    |Banana lemon orange|  1000|    USA|BANANA LEMON ORANGE|
    |Carrots Blueberries|  1500|    USA|CARROTS BLUEBERRIES|
    |              Beans|  1600|    USA|              BEANS|

    In this example, the column `product` is converted to `product_upper` and the contents of this column are converted
    to upper case.
    """

    def func(self, column: Column):
        return upper(column)


class TitleCase(LowerCase):
    """
    This function makes the contents of a column title case.
    This means that every word starts with an upper case.

    Wraps the `pyspark.sql.functions.initcap` function.

    Warnings
    --------
    If the type of the column is not string, TitleCase will not be run. A Warning will be thrown indicating this.

    Parameters
    ----------
    columns : ListOfColumns
        The name of the column or columns to convert to title case. Alias: column.
        Title case will be applied to all columns in the list.
        Column is required to be of string type.
    target_column : str, optional
        The name of the column to store the result in. If None, the result will be stored in the same column as the
        input.

    Example
    -------
    __input_df:__

    |product            |amount|country|
    |-------------------|------|-------|
    |Banana lemon orange|  1000|    USA|
    |Carrots blueberries|  1500|    USA|
    |              Beans|  1600|    USA|

    ```python
    output_df = TitleCase(
        column="product", target_column="product_title"
    ).transform(df)
    ```

    __output_df:__

    |product            |amount|country|product_title      |
    |-------------------|------|-------|-------------------|
    |Banana lemon orange|  1000|    USA|Banana Lemon Orange|
    |Carrots blueberries|  1500|    USA|Carrots Blueberries|
    |              Beans|  1600|    USA|              Beans|

    In this example, the column `product` is converted to `product_title` and the contents of this column are converted
    to title case (each word now starts with an upper case).
    """

    def func(self, column: Column):
        return initcap(column)


# Alias for TitleCase
InitCap = TitleCase



"""
Module for hashing data using SHA-2 family of hash functions

See the docstring of the Sha2Hash class for more information.
"""

from typing import List, Literal, Optional

from pyspark.sql import Column
from pyspark.sql.functions import col, concat_ws, sha2

from koheesio.models import Field
from koheesio.spark.transformations import ColumnsTransformation
from koheesio.spark.utils import SparkDatatype

HASH_ALGORITHM = Literal[224, 256, 384, 512]
STRING = SparkDatatype.STRING


def sha2_hash(columns: List[str], delimiter: Optional[str] = "|", num_bits: Optional[HASH_ALGORITHM] = 256):
    """
    hash the value of 1 or more columns using SHA-2 family of hash functions

    Mild wrapper around pyspark.sql.functions.sha2

    - https://spark.apache.org/docs/3.3.2/api/python/reference/api/pyspark.sql.functions.sha2.html

    Returns the hex string result of SHA-2 family of hash functions (SHA-224, SHA-256, SHA-384, and SHA-512).
    This function allows concatenating the values of multiple columns together prior to hashing.

    If a null is passed, the result will also be null.

    Parameters
    ----------
    columns : List[str]
        The columns to hash
    delimiter : Optional[str], optional, default=|
        Optional separator for the string that will eventually be hashed. Defaults to '|'
    num_bits : Optional[HASH_ALGORITHM], optional, default=256
        Algorithm to use for sha2 hash. Defaults to 256. Should be one of 224, 256, 384, 512
    """
    # make sure all columns are of type pyspark.sql.Column and cast to string
    _columns = []
    for c in columns:
        if isinstance(c, str):
            c: Column = col(c)
        _columns.append(c.cast(STRING.spark_type()))

    # concatenate columns if more than 1 column is provided
    if len(_columns) > 1:
        column = concat_ws(delimiter, *_columns)
    else:
        column = _columns[0]

    return sha2(column, num_bits)


class Sha2Hash(ColumnsTransformation):
    """
    hash the value of 1 or more columns using SHA-2 family of hash functions

    Mild wrapper around pyspark.sql.functions.sha2

    - https://spark.apache.org/docs/3.3.2/api/python/reference/api/pyspark.sql.functions.sha2.html

    Returns the hex string result of SHA-2 family of hash functions (SHA-224, SHA-256, SHA-384, and SHA-512).

    Note
    ----
    This function allows concatenating the values of multiple columns together prior to hashing.

    Parameters
    ----------
    columns : Union[str, List[str]]
        The column (or list of columns) to hash. Alias: column
    delimiter : Optional[str], optional, default=|
        Optional separator for the string that will eventually be hashed. Defaults to '|'
    num_bits : Optional[HASH_ALGORITHM], optional, default=256
        Algorithm to use for sha2 hash. Defaults to 256. Should be one of 224, 256, 384, 512
    target_column : str
        The generated hash will be written to the column name specified here
    """

    delimiter: Optional[str] = Field(
        default="|",
        description="Optional separator for the string that will eventually be hashed. Defaults to '|'",
    )
    num_bits: Optional[HASH_ALGORITHM] = Field(
        default=256, description="Algorithm to use for sha2 hash. Defaults to 256. Should be one of 224, 256, 384, 512"
    )
    target_column: str = Field(
        default=..., description="The generated hash will be written to the column name specified here"
    )

    def execute(self):
        columns = list(self.get_columns())
        self.output.df = (
            self.df.withColumn(
                self.target_column, sha2_hash(columns=columns, delimiter=self.delimiter, num_bits=self.num_bits)
            )
            if columns
            else self.df
        )



"""Loggers are used to log messages from your application.

For a comprehensive guide on the usage, examples, and additional features of the logging classes, please refer to the
[reference/concepts/logger](../reference/concepts/logger.md) section of the Koheesio documentation.

Classes
-------
LoggingFactory
    Logging factory to be used to generate logger instances.
Masked
    Represents a masked value.
MaskedString
    Represents a masked string value.
MaskedInt
    Represents a masked integer value.
MaskedFloat
    Represents a masked float value.
MaskedDict
    Represents a masked dictionary value.
LoggerIDFilter
    Filter which injects run_id information into the log.

Functions
---------
warn
    Issue a warning.

"""

from __future__ import annotations

import inspect
import logging
import os
import sys
from typing import Any, Dict, Generator, Generic, List, Optional, Tuple, TypeVar
from logging import Formatter, Logger, getLogger
from uuid import uuid4
from warnings import warn

from koheesio.utils import import_class

T = TypeVar("T")

__all__ = [
    "Masked",
    "MaskedString",
    "MaskedInt",
    "MaskedFloat",
    "MaskedDict",
    "Logger",
    "LoggerIDFilter",
    "LoggingFactory",
    "warn",
]


class Masked(Generic[T]):
    """
    Represents a masked value.

    Parameters
    ----------
    value : T
        The value to be masked.

    Attributes
    ----------
    _value : T
        The original value.

    Methods
    -------
    __repr__() -> str
        Returns a string representation of the masked value.
    __str__() -> str
        Returns a string representation of the masked value.
    __get_validators__() -> Generator
        Returns a generator of validators for the masked value.
    validate(v: Any, values)
        Validates the masked value.

    """

    def __init__(self, value: T):
        self._value = value

    def __repr__(self) -> str:
        used_as_output = [
            line
            for f in inspect.stack()
            for line in (f.code_context or [])
            if any(e in line.strip() for e in [".debug(", ".info(", ".log(", ".warning(", "print("])
        ]
        return repr(self._value if not used_as_output else "*" * len(str(self._value)) + "(Masked)")

    def __str__(self) -> str:
        return self.__repr__()

    @classmethod
    def __get_validators__(cls) -> Generator:
        """
        Returns a generator that yields the validator function for the class.

        Yields:
            Generator: The validator function for the class.
        """
        yield cls.validate

    @classmethod
    def validate(cls, v: Any, _values):
        """
        Validate the input value and return an instance of the class.

        Parameters
        ----------
        v : Any
            The input value to validate.
        _values : Any
            Additional values used for validation.

        Returns
        -------
        instance : cls
            An instance of the class.

        """
        return cls(v)


class MaskedString(Masked, str):
    """
    Represents a masked string value.
    """

    ...


class MaskedInt(Masked, int):
    """
    Represents a masked integer value.
    """

    ...


class MaskedFloat(Masked, float):
    """
    Represents a masked float value.
    """

    ...


class MaskedDict(Masked, dict):
    """
    Represents a masked dictionary value.
    """

    ...


class LoggerIDFilter(logging.Filter):
    """Filter which injects run_id information into the log."""

    LOGGER_ID: str = str(uuid4())

    def filter(self, record):
        record.logger_id = LoggerIDFilter.LOGGER_ID

        return True


class LoggingFactory:
    """Logging factory to be used to generate logger instances."""

    LOGGER: Optional[logging.Logger] = None
    LOGGER_FILTER: Optional[logging.Filter] = None
    LOGGER_NAME: str = "koheesio"
    LOGGER_LEVEL: str = os.environ.get("KOHEESIO_LOGGING_LEVEL", "WARNING")
    LOGGER_ENV: str = "local"
    LOGGER_FORMAT: str = (
        "[%(logger_id)s] [%(asctime)s] [%(levelname)s] [%(name)s] "
        "{%(module)s.py:%(funcName)s:%(lineno)d} - %(message)s"
    )
    LOGGER_FORMATTER: Formatter = Formatter(LOGGER_FORMAT)
    CONSOLE_HANDLER: Optional[logging.Handler] = None
    ENV: Optional[str] = None

    def __init__(
        self,
        name: Optional[str] = None,
        env: Optional[str] = None,
        level: Optional[str] = None,
        logger_id: Optional[str] = None,
    ):
        """Logging factory to be used in pipeline.Prepare logger instance.

        Parameters
        ----------
        name logger name.
        env environment ("local", "qa", "prod).
        logger_id unique identifier for the logger.
        """

        LoggingFactory.LOGGER_NAME = name or LoggingFactory.LOGGER_NAME
        LoggerIDFilter.LOGGER_ID = logger_id or LoggerIDFilter.LOGGER_ID
        LoggingFactory.LOGGER_FILTER = LoggingFactory.LOGGER_FILTER or LoggerIDFilter()
        LoggingFactory.ENV = env or LoggingFactory.ENV

        console_handler = logging.StreamHandler(sys.stdout if LoggingFactory.ENV == "local" else sys.stderr)
        console_handler.setFormatter(LoggingFactory.LOGGER_FORMATTER)
        console_handler.addFilter(LoggingFactory.LOGGER_FILTER)
        # WARNING is default level for root logger in python
        logging.basicConfig(level=logging.WARNING, handlers=[console_handler], force=True)

        LoggingFactory.CONSOLE_HANDLER = console_handler

        logger = getLogger(LoggingFactory.LOGGER_NAME)
        logger.setLevel(level or LoggingFactory.LOGGER_LEVEL)
        LoggingFactory.LOGGER = logger

    @classmethod
    def __check_koheesio_logger_initialized(cls) -> None:
        """Check that koheesio logger has been initialized"""
        if not cls.LOGGER:
            cls(name=cls.LOGGER_NAME, env=cls.LOGGER_ENV, level=cls.LOGGER_LEVEL, logger_id=LoggerIDFilter.LOGGER_ID)

    @staticmethod
    def add_handlers(handlers: List[Tuple[str, Dict]]) -> None:
        """Add handlers to existing root logger.

        Parameters
        ----------
        handler_class handler module and class for importing.
        handlers_config configuration for handler.

        """
        for handler_module_class, handler_conf in handlers:
            handler_class: logging.Handler = import_class(handler_module_class)
            handler_level = handler_conf.pop("level") if "level" in handler_conf else "WARNING"
            # noinspection PyCallingNonCallable
            handler = handler_class(**handler_conf)
            handler.setLevel(handler_level)
            handler.addFilter(LoggingFactory.LOGGER_FILTER)
            handler.setFormatter(LoggingFactory.LOGGER_FORMATTER)
            LoggingFactory.LOGGER.addHandler(handler)

    @staticmethod
    def get_logger(name: str, inherit_from_koheesio: bool = False) -> Logger:
        """Provide logger. If inherit_from_koheesio then inherit from LoggingFactory.PIPELINE_LOGGER_NAME.

        Parameters
        ----------
        name: Name of logger.
        inherit_from_koheesio: Inherit logger from koheesio

        Returns
        -------
        logger: Logger

        """
        if inherit_from_koheesio:
            LoggingFactory.__check_koheesio_logger_initialized()
            name = f"{LoggingFactory.LOGGER_NAME}.{name}"

        return getLogger(name)



"""
Class for converting DataFrame column names from camel case to snake case.
"""

import re
from typing import Optional

from koheesio.models import Field, ListOfColumns
from koheesio.spark.transformations import ColumnsTransformation

camel_to_snake_re = re.compile("([a-z0-9])([A-Z])")


def convert_camel_to_snake(name: str):
    """
    Converts a string from camelCase to snake_case.

    Parameters:
    ----------
    name : str
        The string to be converted.

    Returns:
    --------
    str
        The converted string in snake_case.
    """
    return camel_to_snake_re.sub(r"\1_\2", name).lower()


class CamelToSnakeTransformation(ColumnsTransformation):
    """
    Converts column names from camel case to snake cases

    Parameters
    ----------
    columns : Optional[ListOfColumns], optional, default=None
        The column or columns to convert. If no columns are specified, all columns will be converted. A list of columns
        or a single column can be specified.
        For example: `["column1", "column2"]` or `"column1"`

    Example
    -------
    __input_df:__

    | camelCaseColumn    | snake_case_column |
    |--------------------|-------------------|
    | ...                | ...               |

    ```python
    output_df = CamelToSnakeTransformation(column="camelCaseColumn").transform(
        input_df
    )
    ```

    __output_df:__

    | camel_case_column | snake_case_column |
    |-------------------|-------------------|
    | ...               | ...               |

    In this example, the column `camelCaseColumn` is converted to `camel_case_column`.

    > Note: the data in the columns is not changed, only the column names.

    """

    columns: Optional[ListOfColumns] = Field(
        default="",
        alias="column",
        description="The column or columns to convert. If no columns are specified, all columns will be converted. "
        "A list of columns or a single column can be specified. For example: `['column1', 'column2']` or `'column1'` ",
    )

    def execute(self):
        _df = self.df

        # Prepare columns input:
        columns = self.df.columns if self.columns == ["*"] else self.columns

        for column in columns:
            _df = _df.withColumnRenamed(column, convert_camel_to_snake(column))

        self.output.df = _df



"""
Koheesio step for running data quality rules with Spark Expectations engine.
"""

from typing import Any, Dict, Optional, Union

from spark_expectations.config.user_config import Constants as user_config
from spark_expectations.core.expectations import (
    SparkExpectations,
    WrappedDataFrameWriter,
)

from pydantic import Field

import pyspark
from pyspark.sql import DataFrame

from koheesio.spark.transformations import Transformation
from koheesio.spark.writers import BatchOutputMode

if pyspark.__version__.startswith("3.5"):
    raise ImportError("Spark Expectations is not supported for Spark 3.5")


class SparkExpectationsTransformation(Transformation):
    """
    Run DQ rules for an input dataframe with Spark Expectations engine.

    References
    ----------
    Spark Expectations: https://engineering.nike.com/spark-expectations/1.0.0/
    """

    product_id: str = Field(default=..., description="Spark Expectations product identifier")
    rules_table: str = Field(default=..., alias="product_rules_table", description="DQ rules table")
    statistics_table: str = Field(default=..., alias="dq_stats_table_name", description="DQ stats table")
    target_table: str = Field(
        default=...,
        alias="target_table_name",
        description="The table that will contain good records. Won't write to it, but will write to the err table with "
        "same name plus _err suffix",
    )
    mode: Union[str, BatchOutputMode] = Field(
        default=BatchOutputMode.APPEND,
        alias="dataframe_writer_mode",
        description="The write mode that will be used to write to the err and stats table. Separate output modes can "
        "be specified for each table using the error_writer_mode and stats_writer_mode params",
    )
    format: str = Field(
        default="delta",
        alias="dataframe_writer_format",
        description="The format used to write to the stats and err table. Separate output formats can be specified for "
        "each table using the error_writer_format and stats_writer_format params",
    )
    error_writing_options: Optional[Dict[str, str]] = Field(
        default_factory=dict, alias="error_writing_options", description="Options for writing to the error table"
    )
    error_writer_mode: Optional[Union[str, BatchOutputMode]] = Field(
        default=BatchOutputMode.APPEND,
        alias="dataframe_writer_mode",
        description="The write mode that will be used to write to the err table",
    )
    error_writer_format: Optional[str] = Field(
        default="delta", alias="dataframe_writer_format", description="The format used to write to the err table"
    )
    stats_writer_mode: Optional[Union[str, BatchOutputMode]] = Field(
        default=BatchOutputMode.APPEND,
        alias="stats_writer_mode",
        description="The write mode that will be used to write to the stats table",
    )
    stats_writer_format: Optional[str] = Field(
        default="delta", alias="stats_writer_format", description="The format used to write to the stats table"
    )
    enable_debugger: bool = Field(default=False, alias="debugger", description="...")
    statistics_streaming: Dict[str, Any] = Field(
        default={user_config.se_enable_streaming: False},
        alias="stats_streaming_options",
        description="SE stats streaming options ",
        validate_default=False,
    )
    se_user_conf: Dict[str, Any] = Field(
        default={
            user_config.se_notifications_enable_email: False,
            user_config.se_notifications_enable_slack: False,
        },
        alias="user_conf",
        description="SE user provided confs",
        validate_default=False,
    )
    drop_meta_column: bool = Field(
        default=False,
        alias="drop_meta_columns",
        description="Whether to drop meta columns added by spark expectations on the output df",
    )

    class Output(Transformation.Output):
        """Output of the SparkExpectationsTransformation step."""

        rules_df: DataFrame = Field(default=..., description="Output dataframe")
        se: SparkExpectations = Field(default=..., description="Spark Expectations object")
        error_table_writer: WrappedDataFrameWriter = Field(
            default=..., description="Spark Expectations error table writer"
        )
        stats_table_writer: WrappedDataFrameWriter = Field(
            default=..., description="Spark Expectations stats table writer"
        )

    @property
    def _error_table_writer(self) -> WrappedDataFrameWriter:
        """Wrapper around Spark DataFrameWriter to be used by Spark Expectations"""
        if not (_writer := self.output.get("error_table_writer", None)):
            _mode = self.error_writer_mode or self.mode
            _format = self.error_writer_format or self.format
            _options = self.error_writing_options
            _writer = WrappedDataFrameWriter().mode(_mode).format(_format).options(**_options)
            self.output.error_table_writer = _writer
        return _writer

    @property
    def _stats_table_writer(self) -> WrappedDataFrameWriter:
        """Wrapper around Spark DataFrameWriter to be used by Spark Expectations"""
        if not (_writer := self.output.get("stats_table_writer", None)):
            _mode = self.stats_writer_mode or self.mode
            _format = self.stats_writer_format or self.format
            _writer = WrappedDataFrameWriter().mode(_mode).format(self.stats_writer_format)
            self.output.stats_table_writer = _writer
        return _writer

    @property
    def _se(self) -> SparkExpectations:
        """Spark Expectations object"""
        if not (_se := self.output.get("se", None)):
            _se = SparkExpectations(
                product_id=self.product_id,
                rules_df=self.output.rules_df,
                stats_table=self.statistics_table,
                stats_table_writer=self._stats_table_writer,
                target_and_error_table_writer=self._error_table_writer,
                debugger=self.enable_debugger,
                stats_streaming_options=self.statistics_streaming,
            )
            self.output.se = _se
        return _se

    def execute(self) -> Output:
        """
        Apply data quality rules to a dataframe using the out-of-the-box SE decorator
        """
        # read rules table
        rules_df = self.spark.read.table(self.rules_table).cache()
        self.output.rules_df = rules_df

        @self._se.with_expectations(
            target_table=self.target_table,
            user_conf=self.se_user_conf,
            # Below params are `False` by default, however exposing them here for extra visibility
            # The writes can be handled by downstream Koheesio steps
            write_to_table=False,
            write_to_temp_table=False,
        )
        def inner(df: DataFrame) -> DataFrame:
            """Just a wrapper to be able to use Spark Expectations decorator"""
            return df

        output_df = inner(self.df)

        if self.drop_meta_column:
            output_df = output_df.drop("meta_dq_run_id", "meta_dq_run_datetime")

        self.output.df = output_df
